"""
Commandline interface
"""
from pathlib import Path
from monty.serialization import loadfn
import numpy as np
import click
from ase.io import read

from easyunfold.unfold import process_projection_options

# pylint:disable=import-outside-toplevel,too-many-locals

DEFAULT_CMAPS = [
    'Purples', 'Greens', 'Oranges', 'Reds', 'Blue', 'YlOrBr', 'YlOrRd', 'OrRd', 'PuRd', 'RdPu', 'BuPu', 'GnBu', 'PuBu', 'YlGnBu', 'PuBuGn',
    'BuGn', 'YlGn'
]


@click.group('easyunfold')
def easyunfold():
    """
    Tool for performing band unfolding
    """
    return


@easyunfold.command()
@click.option('--time-reversal/--no-time-reversal', default=True)
@click.argument('pc-file')
@click.argument('sc-file')
@click.argument('kpoints')
@click.option('--matrix', '-m', help='Transformation matrix')
@click.option('--symprec', help='Transformation matrix', type=float, default=1e-5)
@click.option('--out-file', default='easyunfold.json', help='Name of the output file')
@click.option('--no-expand', help='Do not expand the kpoints by symmetry', default=False, is_flag=True)
@click.option('--nk-per-split', help='Number of band structure kpoints per split.', type=int)
@click.option('--scf-kpoints',
              help='File (IBZKPT) to provide SCF kpoints for self-consistent calculations. Needed for hybrid functional calculations.',
              type=click.Path(exists=True, dir_okay=False))
def generate(pc_file, sc_file, matrix, kpoints, time_reversal, out_file, no_expand, symprec, nk_per_split, scf_kpoints):
    """
    Generate the kpoints for sampling the supercell
    """
    from easyunfold.unfold import UnfoldKSet, read_kpoints

    primitive = read(pc_file)
    supercell = read(sc_file)
    if matrix:
        transform_matrix = matrix_from_string(matrix)
        if not np.allclose(primitive.cell @ transform_matrix, supercell.cell):
            click.echo('Warning: the super cell and the the primitive cell are not commensure.')
            click.echo('Proceed with the assumed tranform matrix')
        click.echo(f'Transform matrix:\n{transform_matrix.tolist()}')
    else:
        tmp = supercell.cell @ np.linalg.inv(primitive.cell)
        transform_matrix = np.rint(tmp)
        if not np.allclose(tmp, transform_matrix):
            click.echo('The super cell and the the primitive cell are not commensure.')
            raise click.Abort()

        click.echo(f'(Guessed) Transform matrix:\n{transform_matrix.tolist()}')

    kpoints, _, labels, _ = read_kpoints(kpoints)
    click.echo(f'{len(kpoints)} kpoints specified along the path')

    unfoldset = UnfoldKSet.from_atoms(transform_matrix,
                                      kpoints,
                                      primitive,
                                      supercell,
                                      time_reversal=time_reversal,
                                      expand=not no_expand,
                                      symprec=symprec)
    unfoldset.kpoint_labels = labels
    try:
        print_symmetry_data(unfoldset)
    except KeyError:
        pass

    out_file = Path(out_file)
    if scf_kpoints is not None:
        scf_kpt, _, _, scf_weights = read_kpoints(scf_kpoints)
        unfoldset.write_sc_kpoints(f'KPOINTS_{out_file.stem}', nk_per_split=nk_per_split, scf_kpoints_and_weights=(scf_kpt, scf_weights))
    else:
        unfoldset.write_sc_kpoints(f'KPOINTS_{out_file.stem}', nk_per_split=nk_per_split)
    click.echo('Supercell kpoints written to KPOITNS_' + out_file.stem)

    # Serialize the data
    Path(out_file).write_text(unfoldset.to_json(), encoding='utf-8')

    click.echo('Unfolding settings written to ' + str(out_file))


@easyunfold.group('unfold')
@click.option('--data-file', default='easyunfold.json', type=click.Path(exists=True, file_okay=True, dir_okay=False))
@click.pass_context
def unfold(ctx, data_file):
    """Perform unfolding and plotting"""

    unfoldset = loadfn(data_file)
    click.echo(f'Loaded data from {data_file}')
    ctx.obj = {'obj': unfoldset, 'fname': data_file}


@unfold.command('status')
@click.pass_context
def unfold_status(ctx):
    """Print the status"""
    from easyunfold.unfold import UnfoldKSet
    unfoldset: UnfoldKSet = ctx.obj['obj']
    print_symmetry_data(unfoldset)
    nkpts_sc = len(unfoldset.expansion_results['reduced_sckpts'])
    click.echo()
    click.echo(f'No. of k points in the primitive cell           : {unfoldset.nkpts_orig}')
    click.echo(f'No. of expanded kpoints to be calculated cell   : {nkpts_sc} ({unfoldset.nkpts_expand})')
    click.echo(f'No. of rotations in the primitive cell          : {unfoldset.pc_opts.shape[0]}')
    click.echo(f'No. of rotations in the super cell              : {unfoldset.sc_opts.shape[0]}')
    click.echo()
    click.echo('Path in the primitive cell:')
    for index, label in unfoldset.kpoint_labels:
        click.echo(f'   {label:<10}: {index+1:<5}')

    if unfoldset.is_calculated:
        click.echo('Unfolding had been performed - use `unfold plot` to plot the spectral function.')
    else:
        click.echo('Please run the supercell band structure calculation and run `unfold calculate`.')


@unfold.command('calculate')
@click.pass_context
@click.argument('wavecar', type=click.Path(exists=True, dir_okay=False), nargs=-1)
@click.option('--save-as')
@click.option('--gamma', is_flag=True)
@click.option('--ncl', is_flag=True)
def unfold_calculate(ctx, wavecar, save_as, gamma, ncl):
    """
    Perform the unfolding

    Multiple WAVECAR files can be supplied for split-path calculations.
    """
    from easyunfold.unfold import UnfoldKSet

    unfoldset: UnfoldKSet = ctx.obj['obj']
    unfoldset.get_spectral_weights(wavecar, gamma, ncl=ncl)

    out_path = save_as if save_as else ctx.obj['fname']
    Path(out_path).write_text(unfoldset.to_json(), encoding='utf-8')
    click.echo('Unfolding data written to ' + out_path)


def add_plot_options(func):
    """Added common plotting options to a function"""
    click.option('--gamma', is_flag=True, help='Is the calculation a gamma only one?')(func)
    click.option('--ncl', is_flag=True, help='Is the calculation with non-colinear spin?')(func)
    click.option('--npoints', type=int, default=2000, help='Number of bins for the energy.', show_default=True)(func)
    click.option('--sigma', type=float, default=0.02, help='Smearing width for the energy in ' 'eV.', show_default=True)(func)
    click.option('--eref', type=float, help='Reference energy in eV.')(func)
    click.option('--emin', type=float, default=-5., help='Minimum energy in eV relative to the ' 'reference.', show_default=True)(func)
    click.option('--emax', type=float, default=5., help='Maximum energy in eV relative to the ' 'reference.', show_default=True)(func)
    click.option('--vscale', type=float, help='A scaling factor for the colour mapping.', default=1.0)(func)
    click.option('--out-file', default='unfold.png', help='Name of the output file.')(func)
    click.option('--cmap',
                 default='PuRd',
                 help='Name of the colour map(s) to use. Passing a list separated by "|" for the '
                 'combined plot.',
                 show_default=True)(func)
    click.option('--show', is_flag=True, default=False, help='Show the plot interactively.')(func)
    click.option('--no-symm-average',
                 is_flag=True,
                 default=False,
                 help='Do not include symmetry '
                 'related kpoints for '
                 'averaging.',
                 show_default=True)(func)
    click.option('--procar', multiple=True, help='PROCAR files used for atomic weighting')(func)
    click.option('--atoms-idx', help='Indices of the atoms to be used for weighting (1-indexed).')(func)
    click.option('--orbitals', help='Orbitals of to be used for weighting.')(func)
    click.option('--title', help='Title to be used')(func)
    return func


@unfold.command('effective-mass')
@click.pass_context
@click.option('--intensity-threshold', type=float, default=0.1, help='Intensity threshold for detecting valid bands.')
@click.option('--spin', type=int, default=0, help='Index of the spin channel.')
@click.option('--npoints', type=int, default=3, help='Number of kpoints used for fitting from the extrema.')
@click.option('--extrema-detect-tol', type=float, default=0.01, help='Tolerance for band extrema detection.')
@click.option('--degeneracy-detect-tol', type=float, default=0.01, help='Tolerance for band degeneracy detection at extrema.')
@click.option('--nocc', type=int)
@click.option('--plot', is_flag=True, default=False)
@click.option('--plot-fit', is_flag=True, default=False, help='Generate plots of the band edge and parabolic fits.')
@click.option('--fit-label', help='Which branch to use for plot fitting. e.g. electrons:0', default='electrons:0')
@click.option('--band-filter', default=None, type=int, help='Only displace information for this band.')
@click.option('--out-file', default='unfold-effective-mass.png', help='Name of the output file.')
def unfold_effective_mass(ctx, intensity_threshold, spin, band_filter, npoints, extrema_detect_tol, degeneracy_detect_tol, nocc, plot,
                          plot_fit, fit_label, out_file):
    """
    Compute and print effective masses by tracing the unfolded weights.

    Note that this functionality only works for simple unfolded band structures,
    and it is likely to fail for complex cases.

    Use the `--plot-fit` flag to generate the detected band edge verses the parabolic fit
    for visual confirmation if unsure.
    """
    from easyunfold.effective_mass import EffectiveMass
    from easyunfold.unfold import UnfoldKSet
    from tabulate import tabulate
    unfoldset: UnfoldKSet = ctx.obj['obj']
    efm = EffectiveMass(unfoldset, intensity_tol=intensity_threshold, extrema_tol=extrema_detect_tol, degeneracy_tol=degeneracy_detect_tol)

    click.echo('Band extrema data:')
    table = []
    for mode in ['cbm', 'vbm']:
        for kid, subkid, iband in zip(*efm.get_band_extrema(mode=mode)):
            band_idx = ','.join(map(str, iband))
            table.append([kid, mode, subkid, band_idx])
    click.echo(tabulate(table, headers=['Kpoint index', 'Kind', 'Sub-kpoint index', 'Band indices']))
    click.echo('')

    if nocc:
        efm.set_nocc(nocc)
    output = efm.get_effective_masses(ispin=spin, npoints=npoints)

    # Filter by band if requested
    if band_filter is not None:
        for carrier in ['electrons', 'holes']:
            output[carrier] = [entry for entry in output[carrier] if entry['band_index'] == band_filter]

    ## Print data
    def print_data(entries, tag='me'):
        """Print the effective mass data"""
        table = []
        for i, entry in enumerate(entries):
            me = entry['effective_mass']

            kf = entry['kpoint_from']
            lf = entry['kpoint_label_from']
            kfrom = f'{kf} ({lf})'

            kt = [round(x, 5) for x in entry['kpoint_to']]
            lt = entry['kpoint_label_to']
            kto = f'{kt} ({lt})'
            bidx = entry['band_index']
            table.append([i, tag, me, bidx, kfrom, kto])
        click.echo(tabulate(table, headers=['index', 'Kind', 'Effective mass', 'Band index', 'from', 'to']))

    click.echo('Electron effective masses:')
    print_data(output['electrons'], 'm_e')
    print('')
    click.echo('Hole effective masses:')
    print_data(output['holes'], 'm_h')

    click.echo('Unfolded band structure can be ambiguous, please cross-check with the spectral function plot.')

    if plot:
        from easyunfold.plotting import UnfoldPlotter
        plotter = UnfoldPlotter(unfoldset)
        click.echo('Generating spectral function plot for visualising detected branches...')
        engs, sf = unfoldset.get_spectral_function()
        plotter.plot_effective_mass(efm, engs, sf, effective_mass_data=output, save=out_file)

    elif plot_fit:
        from easyunfold.plotting import UnfoldPlotter
        plotter = UnfoldPlotter(unfoldset)
        carrier, idx = fit_label.split(':')
        fname = Path(out_file).stem
        ext = Path(out_file).suffix
        for carrier in ['electrons', 'holes']:
            for idx, _ in enumerate(output[carrier]):
                plotter.plot_effective_mass_fit(efm=efm,
                                                npoints=npoints,
                                                carrier=carrier,
                                                idx=int(idx),
                                                save=fname + f'_fit_{carrier}_{idx}' + ext)


@unfold.command('plot')
@click.pass_context
@add_plot_options
def unfold_plot(ctx, gamma, npoints, sigma, eref, out_file, show, emin, emax, cmap, ncl, no_symm_average, vscale, procar, atoms_idx,
                orbitals, title):
    """
    Plot the spectral function

    This command uses the stored unfolding data to plot the effective bands structure (EBS).
    """
    _unfold_plot(ctx, gamma, npoints, sigma, eref, out_file, show, emin, emax, cmap, ncl, no_symm_average, vscale, procar, atoms_idx,
                 orbitals, title)


@unfold.command('plot-projections')
@click.pass_context
@add_plot_options
@click.option('--combined/--no-combined', is_flag=True, default=False, help='Plot all projections in a combined graph.')
@click.option('--intensity', default=1.0, help='Color intensity for combined plot', type=float, show_default=True)
@click.option('--colors', help='Colors to be used for combined plot. Comma separated. Default = "r,g,b".')
def unfold_plot_projections(ctx, gamma, npoints, sigma, eref, out_file, show, emin, emax, cmap, ncl, no_symm_average, vscale, procar,
                            atoms_idx, orbitals, title, combined, intensity, colors):
    """
    Plot with subplots with multiple atomic projections
    """
    from easyunfold.unfold import UnfoldKSet
    from easyunfold.plotting import UnfoldPlotter
    import matplotlib.pyplot as plt

    unfoldset: UnfoldKSet = ctx.obj['obj']
    click.echo(f'Loading projections from: {procar}')

    plotter = UnfoldPlotter(unfoldset)
    fig = plotter.plot_projected(procar,
                                 gamma=gamma,
                                 npoints=npoints,
                                 sigma=sigma,
                                 eref=eref,
                                 ylim=(emin, emax),
                                 cmap=cmap,
                                 ncl=ncl,
                                 symm_average=not no_symm_average,
                                 atoms_idx=atoms_idx,
                                 orbitals=orbitals,
                                 title=title,
                                 vscale=vscale,
                                 use_subplot=not combined,
                                 intensity=intensity,
                                 colors=colors.split(',') if colors is not None else None)

    if out_file:
        fig.savefig(out_file, dpi=300)
        click.echo(f'Unfolded band structure saved to {out_file}')

    if show:
        plt.show()


def _unfold_plot(ctx,
                 gamma,
                 npoints,
                 sigma,
                 eref,
                 out_file,
                 show,
                 emin,
                 emax,
                 cmap,
                 ncl,
                 no_symm_average,
                 vscale,
                 procar,
                 atoms_idx,
                 orbitals,
                 title,
                 ax=None):
    """
    Plot the spectral function

    This command uses the stored unfolding data to plot the effective bands structure (EBS).
    """
    from easyunfold.unfold import UnfoldKSet
    from easyunfold.plotting import UnfoldPlotter

    unfoldset: UnfoldKSet = ctx.obj['obj']
    if not unfoldset.is_calculated:
        click.echo('Unfolding has not been performed yet, please run `unfold calculate` command.')
        raise click.Abort()

    if eref is None:
        eref = unfoldset.calculated_quantities.get('vbm', 0.0)
    click.echo(f'Using a reference energy of {eref:.3f} eV')

    # Process the PROCAR
    if procar:
        click.echo(f'Loading projections from: {procar}')
        unfoldset.load_procar(procar)

    # Setup the atoms_idx and orbitals
    if atoms_idx:
        atoms_idx, orbitals = process_projection_options(atoms_idx, orbitals)
    else:
        atoms_idx = None
        orbitals = None

    eng, spectral_function = unfoldset.get_spectral_function(gamma=gamma,
                                                             npoints=npoints,
                                                             sigma=sigma,
                                                             ncl=ncl,
                                                             atoms_idx=atoms_idx,
                                                             orbitals=orbitals,
                                                             symm_average=not no_symm_average)
    if emin is None:
        emin = eng.min() - eref
    if emax is None:
        emax = eng.max() - eref

    plotter = UnfoldPlotter(unfoldset)
    fig = plotter.plot_spectral_function(eng,
                                         spectral_function,
                                         eref=eref,
                                         save=out_file,
                                         show=False,
                                         ylim=(emin, emax),
                                         vscale=vscale,
                                         cmap=cmap,
                                         title=title,
                                         ax=ax)

    if out_file:
        click.echo(f'Unfolded band structure saved to {out_file}')

    if show:
        fig.show()


def print_symmetry_data(kset):
    """Print the symmetry information"""

    # Print space group information
    sc_spg = kset.metadata['symmetry_dataset_sc']
    click.echo('Supercell cell information:')
    click.echo(' ' * 8 + f'Space group number: {sc_spg["number"]}')
    click.echo(' ' * 8 + f'Internation symbol: {sc_spg["international"]}')
    click.echo(' ' * 8 + f'Point group: {sc_spg["pointgroup"]}')

    pc_spg = kset.metadata['symmetry_dataset_pc']
    click.echo('\nPrimitive cell information:')
    click.echo(' ' * 8 + f'Space group number: {pc_spg["number"]}')
    click.echo(' ' * 8 + f'Internation symbol: {pc_spg["international"]}')
    click.echo(' ' * 8 + f'Point group: {pc_spg["pointgroup"]}')


def matrix_from_string(string):
    """Parse transform matrix from a string"""
    elems = [float(x) for x in string.split()]
    # Try gussing the transform matrix
    if len(elems) == 3:
        transform_matrix = np.diag(elems)
    else:
        transform_matrix = np.array(elems).reshape((3, 3))
    return transform_matrix



"""
Code for reading the eigenvalues and plane wave coefficients from the WAVECAR file.
"""

import os
from math import sqrt

import numpy as np
from .vasp_constant import RYTOEV, TPI, AUTOA, HSQDTM


class Wavecar:  # pylint: disable=too-many-instance-attributes
    """
    Class for processing VASP Pseudowavefunction stored in WAVECAR.

    This class is a trimmed-down from that of [PyVaspwfc](https://github.com/QijingZheng/VaspBandUnfolding)
    by QijingZheng <zqj.kaka@gmail.com>.

    Only the functions needed for band unfolding remains.
    The original class has many other functionalities such as ELF calculation.

    The format of VASP WAVECAR, as shown in
        http://www.andrew.cmu.edu/user/feenstra/wavetrans/
    is:
    ```
        Record-length #spin components RTAG(a value specifying the precision)
        #k-points #bands ENCUT(maximum energy for plane waves)
        LatVec-A
        LatVec-B
        LatVec-C
        Loop over spin
           Loop over k-points
              #plane waves, k vector
              Loop over bands
                 band energy, band occupation
              End loop over bands
              Loop over bands
                 Loop over plane waves
                    Plane-wave coefficient
                 End loop over plane waves
              End loop over bands
           End loop over k-points
        End loop over spin
    ```
    """

    def __init__(self, fnm='WAVECAR', lsorbit=False, lgamma=False, gamma_half='x'):
        r"""
        Initialization.

        :param fnm: Path to the `WAVECAR` file
        :param lsorbit: Whether the `WAVECAR` is from a spint-orbit coupling calculation
        :param lgamma: Wether the `WAVECAR` is from a $\Gamma$-only calculation
        """

        self._fname = fnm
        # the directory containing the input file
        self._dname = os.path.dirname(fnm)
        if self._dname == '':
            self._dname = '.'

        self._lsoc = lsorbit
        self._lgam = lgamma
        self._gam_half = gamma_half.lower()

        assert not (lsorbit and lgamma), 'Cannot have both `lsorbit` and `lgamma`!'
        assert self._gam_half in ['x', 'z'], 'Gamma_half must be "x" or "z"'

        self._wfc = open(self._fname, 'rb')  # pylint:disable=consider-using-with

        # read the basic information
        self.read_wf_header()
        # read the band information
        self.read_bands()

        if self._lsoc:
            assert self._nspin == 1, 'NSPIN = 1 for noncollinear version WAVECAR!'

    def close(self):
        """Close the file handle"""
        self._wfc.close()

    def is_soc(self):
        """
        Return `True` if the WAVECAR from an SOC calculation.
        """
        return bool(self._lsoc)

    def is_gamma(self):
        r"""
        Return `True` is the WAVECAR is from an $\Gamma$-only calculation.
        """
        return bool(self._lgam)

    def read_wf_header(self):
        """
        Read the system information from WAVECAR, which is written in the first
        two record.

        :rec1:
            `recl, nspin, rtag`
        :rec2:
            `nkpts, nbands, encut, ((cell(i,j) i=1, 3), j=1, 3)`
        """

        # goto the start of the file and read the first record
        self._wfc.seek(0)
        self._recl, self._nspin, self._rtag = np.array(np.fromfile(self._wfc, dtype=np.float64, count=3), dtype=np.int64)
        self._precision = self.set_wf_prec()
        # the second record
        self._wfc.seek(self._recl)
        dump = np.fromfile(self._wfc, dtype=np.float64, count=12)

        self._nkpts = int(dump[0])  # No. of k-points
        self._nbands = int(dump[1])  # No. of bands
        self._encut = dump[2]  # Energy cutoff
        # real space supercell basis
        self._realspace_cell = dump[3:].reshape((3, 3))
        # real space supercell volume
        self._realspace_cell_volume = np.linalg.det(self._realspace_cell)
        # reciprocal space supercell volume
        self._reciprocal_cell_volume = np.linalg.inv(self._realspace_cell).T

        # Minimum FFT grid size
        cell_abc = np.linalg.norm(self._realspace_cell, axis=1)
        cut_off = np.ceil(sqrt(self._encut / RYTOEV) / (TPI / (cell_abc / AUTOA)))
        self._ngrid = np.array(2 * cut_off + 1, dtype=int)

    def set_wf_prec(self):
        """
        Set wavefunction coefficients precision:

        :TAG = 45200:
            single precision complex, np.complex64, or complex(qs)
        :TAG = 45210:
            double precision complex, np.complex128, or complex(q)
        """
        if self._rtag == 45200:
            return np.complex64
        if self._rtag == 45210:
            return np.complex128
        if self._rtag == 53300:
            raise ValueError('VASP5 WAVECAR format, not implemented yet')
        if self._rtag == 53310:
            raise ValueError('VASP5 WAVECAR format with double precision ' + 'coefficients, not implemented yet')
        raise ValueError(f'Invalid TAG values: {self._rtag}')

    def read_bands(self) -> tuple:
        """
        Extract KS energies and Fermi occupations from WAVECAR.
        """

        self._nplws = np.zeros(self._nkpts, dtype=int)
        self._kvecs = np.zeros((self._nkpts, 3), dtype=float)
        self._bands = np.zeros((self._nspin, self._nkpts, self._nbands), dtype=float)
        self._occs = np.zeros((self._nspin, self._nkpts, self._nbands), dtype=float)

        for ii in range(self._nspin):
            for jj in range(self._nkpts):
                rec = self.locate_rec(ii + 1, jj + 1, 1) - 1
                self._wfc.seek(rec * self._recl)
                dump = np.fromfile(self._wfc, dtype=np.float64, count=4 + 3 * self._nbands)
                if ii == 0:
                    self._nplws[jj] = int(dump[0])
                    self._kvecs[jj] = dump[1:4]
                dump = dump[4:].reshape((-1, 3))
                self._bands[ii, jj, :] = dump[:, 0]
                self._occs[ii, jj, :] = dump[:, 2]

        if self._nkpts > 1:
            tmp = np.linalg.norm(np.dot(np.diff(self._kvecs, axis=0), self._reciprocal_cell_volume), axis=1)
            self._kpath = np.concatenate(([
                0,
            ], np.cumsum(tmp)))
        else:
            self._kpath = None
        return self._kpath, self._bands

    def get_gvectors(self, ikpt=1, force_gamma=False, check_consistency=True):
        r"""
        Generate the G-vectors that satisfies the following relation

        $$
            \frac{|\vec{G} + \vec{k}|^2}{2} < E_{cut}
        $$

        """
        assert 1 <= ikpt <= self._nkpts, 'Invalid kpoint index!'

        kvec = self._kvecs[ikpt - 1]
        # force_gamma: consider gamma-only case regardless of the actual setting
        lgam = True if force_gamma else self._lgam

        fx, fy, fz = [np.arange(n, dtype=int) for n in self._ngrid]
        fx[self._ngrid[0] // 2 + 1:] -= self._ngrid[0]
        fy[self._ngrid[1] // 2 + 1:] -= self._ngrid[1]
        fz[self._ngrid[2] // 2 + 1:] -= self._ngrid[2]
        if lgam:
            if self._gam_half == 'x':
                fx = fx[:self._ngrid[0] // 2 + 1]
            else:
                fz = fz[:self._ngrid[2] // 2 + 1]

    # In meshgrid, fx run the fastest, fz the slowest
        gz, gy, gx = np.array(np.meshgrid(fz, fy, fx, indexing='ij')).reshape((3, -1))
        kgrid = np.array([gx, gy, gz], dtype=float).T
        if lgam:
            if self._gam_half == 'z':
                kgrid = kgrid[(gz > 0) | ((gz == 0) & (gy > 0)) | ((gz == 0) & (gy == 0) & (gx >= 0))]
            else:
                kgrid = kgrid[(gx > 0) | ((gx == 0) & (gy > 0)) | ((gx == 0) & (gy == 0) & (gz >= 0))]

        # Kinetic_Energy = (G + k)**2 / 2
        kinetic_energy = HSQDTM * np.linalg.norm(np.dot(kgrid + kvec[np.newaxis, :], TPI * self._reciprocal_cell_volume), axis=1)**2
        # find Gvectors where (G + k)**2 / 2 < ENCUT
        g_vectors = kgrid[np.where(kinetic_energy < self._encut)[0]]

        # Check if the calculated number of planewaves and the one recorded in the
        # WAVECAR are equal
        if check_consistency:

            if g_vectors.shape[0] != self._nplws[ikpt - 1]:
                if g_vectors.shape[0] * 2 == self._nplws[ikpt - 1]:
                    if not self._lsoc:
                        raise ValueError("""
                        It seems that you are reading a WAVECAR from a NONCOLLINEAR VASP.
                        Please set 'lsorbit = True' when loading the WAVECAR.
                        For example:

                            wfc = Wavecar('WAVECAR', lsorbit=True)
                        """)
                elif g_vectors.shape[0] == 2 * self._nplws[ikpt - 1] - 1:
                    if not self._lgam:
                        raise ValueError("""
                        It seems that you are reading a WAVECAR from a GAMMA-ONLY VASP.  Please set
                        'lgamma = True' when loading the WAVECAR.  Moreover, you may want to set
                        "gamma_half" if you are using VASP version <= 5.2.x.  For VASP <= 5.2.x, check
                        which FFT VASP uses by the following command:

                            $ grep 'use.* FFT for wave' OUTCAR

                        Then

                            # for parallel FFT, VASP <= 5.2.x
                            wfc = Wavecar('WAVECAR', lgamma=True, gamma_half='z')

                            # for serial FFT, VASP <= 5.2.x
                            wfc = Wavecar('WAVECAR', lgamma=True, gamma_half='x')

                        For VASP >= 5.4, WAVECAR is written with x-direction half grid regardless of
                        parallel or serial FFT.

                            # "gamma_half" default to "x" for VASP >= 5.4
                            wfc = Wavecar('WAVECAR', lgamma=True, gamma_half='x')
                        """)
                else:
                    raise ValueError(f"""
                    NO. OF PLANEWAVES NOT CONSISTENT:

                        THIS CODE -> {g_vectors.shape[0]}
                        FROM VASP -> {self._nplws[ikpt - 1] // 2 if self._lsoc else self._nplws[ikpt - 1]}
                           NGRIDS -> {np.prod(self._ngrid)}
                    """)

        return np.asarray(g_vectors, dtype=int)

    def read_band_coeffs(self, ispin=1, ikpt=1, iband=1, norm=False):
        """
        Read the planewave coefficients of specified KS states.
        """

        self.check_index(ispin, ikpt, iband)

        rec = self.locate_rec(ispin, ikpt, iband)
        self._wfc.seek(rec * self._recl)

        nplw = self._nplws[ikpt - 1]
        dump = np.fromfile(self._wfc, dtype=self._precision, count=nplw)

        cg = np.asarray(dump, dtype=np.complex128)
        if norm:
            cg /= np.linalg.norm(cg)
        return cg

    def locate_rec(self, ispin=1, ikpt=1, iband=1):
        """
        Return the rec position for specified KS state.
        """

        self.check_index(ispin, ikpt, iband)

        rec = 2 + (ispin - 1) * self._nkpts * (self._nbands + 1) + \
                  (ikpt - 1) * (self._nbands + 1) + \
            iband
        return rec

    def check_index(self, ispin, ikpt, iband):
        """
        Check if the index is valid!
        """
        assert 1 <= ispin <= self._nspin, 'Invalid spin index!'
        assert 1 <= ikpt <= self._nkpts, 'Invalid kpoint index!'
        assert 1 <= iband <= self._nbands, 'Invalid band index!'



"""
Module for obtaining effective mass
"""
from typing import Union

import numpy as np
from scipy.constants import physical_constants
from scipy.optimize import curve_fit

from .unfold import UnfoldKSet
# pylint: disable=invalid-name

eV_to_hartree = physical_constants['electron volt-hartree relationship'][0]
bohr_to_m = physical_constants['Bohr radius'][0]
angstrom_to_bohr = bohr_to_m / 1e-10

TMP_DATA = {}


def fit_effective_mass(distances, energies, parabolic=True):
    """Fit the effective masses using either a parabolic or nonparabolic fit.

    Adapted from ``sumo``.

    Args:
        distances (:obj:`numpy.ndarray`): The x-distances between k-points in
            reciprocal Angstroms, normalised to the band extrema.
        energies (:obj:`numpy.ndarray`): The band eigenvalues normalised to the
            eigenvalue of the band extrema.
        parabolic (:obj:`bool`, optional): Use a parabolic fit of the band
            edges. If ``False`` then nonparabolic fitting will be attempted.
            Defaults to ``True``.

    Returns:
        float: The effective mass in units of electron rest mass, :math:`m_0`.
    """
    if parabolic:
        fit = np.polyfit(distances, energies, 2)
        c = 2 * fit[0]  # curvature therefore 2 * the exponent on the ^2 term

    else:
        # Use non parabolic description of the bands
        def f(x, alpha, d):
            top = np.sqrt(4 * alpha * d * x**2 + 1) - 1
            bot = 2 * alpha
            return top / bot

        # set boundaries for curve fitting: alpha > 1e-8
        # as alpha = 0 causes an error
        bounds = ((1e-8, -np.inf), (np.inf, np.inf))
        popt, _ = curve_fit(f, distances, energies, p0=[1.0, 1.0], bounds=bounds)  # pylint: disable=unbalanced-tuple-unpacking
        c = 2 * popt[1]

    # coefficient is currently in eV/Angstrom^2/h_bar^2
    # want it in atomic units so Hartree/bohr^2/h_bar^2
    eff_mass = (angstrom_to_bohr**2 / eV_to_hartree) / c
    return eff_mass


def fitted_band(x: np.ndarray, eff_mass: float) -> np.ndarray:
    """Return fitted effective mass curve"""
    c = (angstrom_to_bohr**2 / eV_to_hartree) / eff_mass
    x0 = x - x[0]
    return x0 + x[0], c / 2 * x0**2


def points_with_tol(array, value, tol=1e-4):
    """
    Return the indices and values of points in an array close to the value with a tolerance
    """
    idx = np.where(np.abs(array - value) < tol)[0]
    return idx, array[idx]


class EffectiveMass:
    """Calculate effective mass from unfolding data"""

    def __init__(
        self,
        unfold: UnfoldKSet,
        intensity_tol: float = 1e-1,
        extrema_tol: float = 1e-3,
        degeneracy_tol: float = 1e-2,
        parabolic: bool = True,
        npoints: float = 3,
    ):
        """
        Instantiate the object

        Args:
            unfold (UnfoldKSet): The ``UnfoldKSet`` object that holes unfolding data.
            intensity_tol (float): Intensity tolerance for detecting band edges.
            parabolic (bool): Perform parabolic fit or not. Defaults to True as non-parabolic fit is not working at the moment...
        """
        self.unfold: UnfoldKSet = unfold
        self.intensity_tol = intensity_tol
        self.extrema_detect_tol = extrema_tol
        self.degeneracy_tol = degeneracy_tol
        self.parabolic = parabolic
        self.nocc = None  # Number of occupied bands
        if npoints < 3:
            raise ValueError('At least three points are needed for fitting the effective mass!')

        self.npoints = npoints

    def set_nocc(self, nocc):
        self.nocc = nocc

    @property
    def kpoints(self):
        return self.unfold.kpts_pc

    @property
    def kpoints_labels(self):
        return self.unfold.kpoint_labels

    def get_band_extrema(self, mode: str = 'cbm', extrema_tol: float = None, degeneracy_tol: float = None, ispin=0):
        """
        Obtain the kpoint idx of band maximum, sub indices in th set and the band indices.

        The search takes two steps, first the kpoints at the band extrema is located by comparing the
        band energies with that recorded in supplied *cbm* and *vbm*, based on the `exgtrema_tol`.
        Afterwards, the band indices are selected at the these kpoints using `degeneracy_tol`.

        Returns:
            A tuple of extrema locations including a list of kpoint indices, sub-indices within
            the set and the band indices at each kpoint that is within the `tol` set.
        """
        if extrema_tol is None:
            extrema_tol = self.extrema_detect_tol
        if degeneracy_tol is None:
            degeneracy_tol = self.degeneracy_tol

        intensity_tol = self.intensity_tol

        if mode not in ['cbm', 'vbm']:
            raise ValueError(f'Unknown mode {mode}')
        cbm = self.unfold.calculated_quantities[mode]
        weights = self.unfold.calculated_quantities['spectral_weights_per_set']

        # Indices of the kpoint corresponding to the CBM
        k_indicies = []
        k_subset_indices = []
        cbm_indices = []
        for ik, wset in enumerate(weights):
            for isubset in range(wset.shape[1]):
                if np.any(np.abs(wset[ispin, isubset, :, 0] - cbm) < extrema_tol):
                    itmp, _ = points_with_tol(wset[ispin, isubset, :, 0], cbm, extrema_tol)
                    # Check if it has sufficient intensity
                    if np.max(wset[ispin, isubset, itmp, 1]) < intensity_tol:
                        continue
                    # Select this kpoints
                    k_indicies.append(ik)
                    # Select all band indices within the tolerance
                    k_subset_indices.append(isubset)
                    # Stop looking at other kpoints in the k subset if found
                    break

        # Go through each case
        for ik, iksub in zip(k_indicies, k_subset_indices):
            itmp, _ = points_with_tol(weights[ik][ispin, iksub, :, 0], cbm, degeneracy_tol)
            cbm_indices.append(itmp)

        return k_indicies, k_subset_indices, cbm_indices

    def _get_kpoint_distances(self):
        """
        Distances between the kpoints along the path in the reciprocal space.
        This does not take account of the breaking of the path
        NOTE: the reciprocal lattice vectors includes the 2pi factor, e.g. np.linalg.inv(L).T * 2 * np.pi
        """
        kpts = self.kpoints
        pc_latt = self.unfold.pc_latt

        kpts_path = kpts @ np.linalg.inv(pc_latt).T * np.pi * 2  # Kpoint path in the reciprocal space

        dists = np.cumsum(np.linalg.norm(np.diff(kpts_path, axis=0), axis=-1))
        dists = np.append([0], dists)
        return dists

    def _get_fitting_data(self, kidx: int, iband: int, direction=1, ispin=0, npoints=None):
        """
        Get fitting data for a specific combination of kpoint and band index
        """
        istart = kidx
        weights = self.unfold.calculated_quantities['spectral_weights_per_set']
        dists = self._get_kpoint_distances()
        kdists = []
        engs_effective = []

        npoints = self.get_npoints(npoints)
        for i in range(npoints):
            idx = istart + i * direction
            kdists.append(dists[idx])
            # Get the spectral weight array
            sw = weights[idx]
            intensities = sw[ispin, :, iband, 1]
            engs = sw[ispin, :, iband, 0]
            kw = self.unfold.expansion_results['weights'][idx]
            # Compute the effective energy weighted by intensity and kpoint weighting
            eng_effective = np.sum(engs * intensities * kw) / np.sum(intensities * kw)
            engs_effective.append(eng_effective)
        return kdists, engs_effective

    def get_npoints(self, override: Union[float, None] = None):
        """Get the number of points used for fitting"""
        if override is None:
            return self.npoints
        return override

    def get_effective_masses(self, npoints: Union[float, None] = None, ispin=0):
        """
        Workout the effective masses based on the unfolded band structure
        """
        outputs = {}
        for mode in ['cbm', 'vbm']:
            name = 'electrons' if mode == 'cbm' else 'holes'
            outputs[name] = self._get_effective_masses(mode, npoints=npoints, ispin=ispin)
        return outputs

    def _get_effective_masses(self, mode: str = 'cbm', ispin: int = 0, npoints: Union[None, int] = None):
        """
        Workout the effective masses based on the unfolded band structure for CBM or VBM
        """
        iks, _, iband = self.get_band_extrema(mode=mode)
        # Override occupations
        if self.nocc:
            iband = [self.nocc for _ in iband]

        npoints = self.get_npoints(npoints)
        results = []
        label_idx = [x[0] for x in self.kpoints_labels]
        label_names = [x[1] for x in self.kpoints_labels]

        for idxk, idxb in zip(iks, iband):
            for direction in (-1, 1):
                # Check if the direction makes sense
                if idxk + direction < 0 or idxk + direction >= len(self.kpoints):
                    continue
                # Check if the direction is broken
                if idxk + direction in label_idx:
                    continue
                # Get fitting data for each (degenerate) band at the extrema
                for band_id in idxb:
                    kdists, engs_effective = self._get_fitting_data(idxk, band_id, direction, ispin=ispin, npoints=npoints)

                    me = fit_effective_mass(kdists, engs_effective, parabolic=self.parabolic)

                    # If the identified edge is not in the list of high symmetry point, ignore it
                    # This mitigate the problem where the CBM can be duplicated....
                    ilabel, label_from, label_to = locate_kpoint_segment(idxk, label_idx, label_names, direction)
                    # Record the results
                    results.append({
                        'kpoint_index': idxk,
                        'direction': direction,
                        'effective_mass': me,
                        'band_index': band_id,
                        'kpoint_label_from': label_from,
                        'kpoint_from': self.kpoints[idxk],
                        'kpoint_label_to': label_to,
                        'kpoint_to': self.kpoints[label_idx[ilabel + direction]],
                        'type': 'electrons' if mode == 'cbm' else 'holes',
                        'raw_data': {
                            'kpoint_distances': kdists,
                            'effective_energies': engs_effective
                        }
                    })
        results.sort(key=lambda x: abs(abs(x['effective_mass'])))
        results.sort(key=lambda x: abs(x['kpoint_index']))
        return results


def locate_kpoint_segment(idxk: int, label_idx: list, label_names: list, direction: int):
    """Locate the labels and indices of the kpoints defining a segment"""
    if idxk not in label_idx:
        pairs = list(zip(label_idx, label_names))
        i = 0
        for i, (ikto, _) in enumerate(pairs):
            if ikto > idxk:
                break
        label_to = label_names[i]
        label_from = 'N/A'
        ilabel = label_idx.index(label_idx[i - 1])
    else:
        ilabel = label_idx.index(idxk)
        label_from = label_names[ilabel]
        label_to = label_names[ilabel + direction]
    return ilabel, label_from, label_to



"""
Procar reader
"""
from typing import List, Union
from pathlib import Path
import re
import numpy as np


class Procar:
    """Reader for PROCAR file"""

    def __init__(self, fobj_or_path=None, is_soc=False):
        """
        Read the PROCAR file from a handle or path

        :param fobj_or_path:  A file-like obj or a path
        :param is_soc: Whether the PROCAR is from a calculation with spin-orbit coupling
        """
        self._is_soc = is_soc
        self.eigenvalues = None
        self.kvecs = None
        self.kweights = None
        self.nbands = None
        self.nkpts = None
        self.nspins = None
        self.nion = None
        self.occs = None
        self.proj_names = None
        self.proj_data = None
        self.header = None
        self.proj_xyz = None

        # Read the PROCAR
        if isinstance(fobj_or_path, (str, Path)):
            with open(fobj_or_path, encoding='utf-8') as fhandle:
                self._read(fhandle)
        else:
            self._read(fobj_or_path)

    def _read(self, fobj):
        """Main function for reading in the data"""

        # First sweep - found the number of kpoints and the number of bands
        fobj.seek(0)
        self.header = fobj.readline()
        # Read the NK, NB and NIONS that are integers
        self.nkpts, self.nbands, self.nion = [int(token) for token in re.sub(r'[^0-9]', ' ', fobj.readline()).split()]
        # Number of projects and their names
        nproj = None
        self.proj_names = None
        for line in fobj:
            if re.match(r'^ion', line):
                nproj = len(line.strip().split()) - 2
                self.proj_names = line.strip().split()[1:-1]
                break
        # Count the number of data lines, these lines do not have any alphabets
        proj_data = []
        energies = []
        occs = []
        kvecs = []
        kweights = []
        fobj.seek(0)
        for line in fobj:
            if not re.search(r'[a-zA-Z]', line) and line.strip():
                proj_data.append([float(token) for token in line.strip().split()[1:-1]])
            elif line.startswith('band'):
                tokens = line.strip().split()
                energies.append(float(tokens[4]))
                occs.append(float(tokens[-1]))
            elif line.startswith(' k-point'):
                line = re.sub(r'(\d)-', r'\1 -', line)
                tokens = line.strip().split()
                kvecs.append([float(val) for val in tokens[-6:-3]])
                kweights.append(float(tokens[-1]))
        self.occs = np.array(occs)
        self.kvecs = np.array(kvecs)
        self.kweights = np.array(kweights)
        self.eigenvalues = np.array(energies)

        proj_data = np.array(proj_data, dtype=float)

        self.nspins = proj_data.shape[0] // (self.nion * self.nbands * self.nkpts)
        self.nspins //= 4 if self._is_soc else 1
        if self._is_soc:
            self.spin = self.spin // 4
        # Reshape
        self.occs.resize((self.nspins, self.nkpts, self.nbands))
        self.kvecs.resize((self.nspins, self.nkpts, 3))
        self.kweights.resize((self.nspins, self.nkpts))
        self.eigenvalues.resize((self.nspins, self.nkpts, self.nbands))

        # Reshape the array
        if self._is_soc is False:
            self.proj_data = proj_data.reshape((self.nspins, self.nkpts, self.nbands, self.nion, nproj))
        else:
            self.proj_data = proj_data.reshape((self.nspins, self.nkpts, self.nbands, 4, self.nion, nproj))
            # Split the data into xyz projection and total
            self.proj_xyz = self.proj_data[:, :, :, 1:, :, :]
            self.proj_data = self.proj_data[:, :, :, 0, :, :]

    def get_projection(self, atom_idx: List[int], proj: Union[List[str], str], weight_by_k=False):
        """
        Get project for specific atoms and specific projectors


        :param atom_idx: A list of index of the atoms to be selected
        :param proj: A list of the projector names to be selected
        :param weight_by_k: Apply k weighting or not.

        :returns: The project summed over the selected atoms and the projectors
        """
        atom_mask = [iatom in atom_idx for iatom in range(self.nion)]
        assert any(atom_mask)
        if proj == 'all':
            out = self.proj_data[:, :, :, atom_mask, :].sum(axis=(-1, -2))
        else:
            proj_idx = [proj_name in proj for proj_name in self.proj_names]
            assert any(proj_idx)
            out = self.proj_data[:, :, :, :, proj_idx]
            out = out[:, :, :, atom_idx, :].sum(axis=(-1, -2))

        if weight_by_k:
            for kidx in range(self.nkpts):
                out[:, kidx, :] *= self.kweights[kidx]
        return out



"""
Plotting utilities
"""
from typing import Union
import matplotlib.pyplot as plt

import numpy as np
from colormath.color_conversions import convert_color
from colormath.color_objects import (
    HSVColor,
    LabColor,
    LCHabColor,
    LCHuvColor,
    XYZColor,
    sRGBColor,
)
from matplotlib.colors import to_rgb

from .unfold import UnfoldKSet, clean_latex_string, process_projection_options
from .effective_mass import EffectiveMass, fitted_band

# pylint: disable=too-many-locals, too-many-arguments


class UnfoldPlotter:
    """A collection plotting tools for unfolded band structures"""

    def __init__(self, unfold: UnfoldKSet):
        """
        Instantiate the plotter object

        :param unfold: `UnfoldKSet` object to be *plotted*.
        """
        self.unfold = unfold

    def plot_spectral_function(
        self,
        engs: np.ndarray,
        sf: np.ndarray,
        eref=None,
        figsize=(4, 3),
        ylim=(-3, 3),
        dpi=150,
        vscale=1.0,
        contour_plot=False,
        alpha=1.0,
        save=False,
        ax=None,
        vmin=None,
        vmax=None,
        cmap='PuRd',
        show=False,
        title=None,
    ):
        """
        Plot the spectral function.

        :param np.ndarray engs: The energies of the spectral functions.
        :param sf: An array of the spectral function.
        :param float eref: Reference energy to be used - this energy will be set as zero.
        :param figsize: Size of the figure.
        :param ylim: Plotting limit for the y-axis, with respect to `eref`.
        :param dpi: DPI of the generated graph.
        :param vscale: A scaling factor for the colour map.
        :param contour_plot): Whether to use contour plot instead of normal meshed color map.
        :param alphathe color map.
        :param savethe file where the generated figure is saved.
        :param ax: An existing plotting axis to be be used.
        :param vminfor the color map.
        :param vmaxfor the color map.
        :param cmap: Name of the color map to be used.

        :returns: The figure generated containing the spectral function.
        """
        unfold = self.unfold
        kdist = unfold.get_kpoint_distances()
        # Extend of x axis
        xmin, xmax = kdist.min(), kdist.max()
        nspin = sf.shape[0]

        if eref is None:
            eref = unfold.calculated_quantities.get('vbm', 0.0)

        if ax is None:
            if nspin == 1:
                fig, ax = plt.subplots(1, 1, figsize=figsize, dpi=dpi)
                axes = [ax]
            else:
                fig, axes = plt.subplots(1, 2, figsize=figsize, dpi=dpi)
        else:
            if not isinstance(ax, list):
                axes = [ax]
            else:
                axes = ax
            fig = axes[0].figure

        # Shift the kdist so the pcolormesh draw the pixel centred on the original point
        X, Y = np.meshgrid(kdist, engs - eref)

        # Calculate the min and max values within the field of view, scaled by the factor
        mask = (engs < (ylim[1] + eref)) & (engs > (ylim[0] + eref))
        vmin = sf[:, mask, :].min() if vmin is None else vmin

        if vmax is None:
            vmax = sf[:, mask, :].max()
            vmax = (vmax - vmin) * vscale + vmin

        for ispin, ax_ in zip(range(nspin), axes):
            if contour_plot:
                ax_.contourf(X, Y, sf[ispin], cmap=cmap, vmax=vmax, vmin=vmin, alpha=alpha)
            else:
                ax_.pcolormesh(X, Y, sf[ispin], cmap=cmap, shading='auto', vmax=vmax, vmin=vmin, alpha=alpha)

            ax_.set_xlim(xmin, xmax)
            ax_.set_ylim(*ylim)
            ax_.set_ylabel('Energy (eV)', labelpad=5)
            if title:
                ax_.set_title(title)

            # Label the kpoints
            self._add_kpoint_labels(ax_)

        fig.tight_layout(pad=0.2)
        if save:
            fig.savefig(save, dpi=300)
        if show:
            fig.show()
        return fig

    def _plot_spectral_function_rgba(
        self,
        engs: np.ndarray,
        sf: np.ndarray,
        eref: Union[None, float] = None,
        figsize=(4, 3),
        ylim=(-3, 3),
        dpi: float = 150,
        intensity: float = 1.0,
        save: bool = False,
        ax: Union[None, plt.Axes] = None,
        show: bool = False,
        title: Union[None, str] = None,
        vmin: Union[None, float] = None,
        vmax: Union[None, float] = None,
    ):
        """
        Plot spectral function defined as RGBA colours

        :param np.ndarray engs: The energies of the spectral functions.
        :param sf: An array of the spectral function.
        :param float eref: Reference energy to be used - this energy will be set as zero.
        :param figsize: Size of the figure.
        :param ylim: Plotting limit for the y-axis, with respect to `eref`.
        :param dpi: DPI of the generated graph.
        :param save: The file where the generated figure is saved.
        :param ax: An existing plotting axis to be be used.
        :param vmin: Lower bound for constructing the alpha channel.
        :param vmax: Upper bound for constructing the alpha channel.
        :param intensity: Scaling factor for the alpha channel.
        :param title: Title for the plot.

        :returns: the figure generated containing the spectral function.
        """
        unfold = self.unfold
        nspin = sf.shape[0]

        if eref is None:
            eref = unfold.calculated_quantities.get('vbm', 0.0)

        if ax is None:
            if nspin == 1:
                fig, ax = plt.subplots(1, 1, figsize=figsize, dpi=dpi)
                axes = [ax]
            else:
                fig, axes = plt.subplots(1, 2, figsize=figsize, dpi=dpi)
        else:
            if not isinstance(ax, list):
                axes = [ax]
            else:
                axes = ax
            fig = axes[0].figure

        mask = (engs < (ylim[1] + eref)) & (engs > (ylim[0] + eref))
        vmin = sf[:, mask, :, 3].min() if vmin is None else vmin

        if vmax is None:
            vmax = sf[:, mask, :, 3].max()

        # Clip the alpha range
        alpha = sf[:, :, :, 3]
        alpha = (alpha - vmin) / (vmax - vmin) * intensity
        alpha = np.clip(alpha, 0, 1)
        sf[:, :, :, 3] = alpha

        for ispin, ax_ in zip(range(nspin), axes):
            ax_.imshow(sf[ispin], extent=[0, sf.shape[2], max(engs) - eref, min(engs) - eref], aspect='auto')
            ax_.set_ylim(ylim)
            ax_.set_ylabel('Energy (eV)', labelpad=5)
            ax_.set_title(title)
            self._add_kpoint_labels(ax_, x_is_kidx=True)

        fig.tight_layout(pad=0.2)
        if save:
            fig.savefig(save, dpi=300)
        if show:
            fig.show()
        return fig

    def _add_kpoint_labels(self, ax: plt.Axes, x_is_kidx=False):
        """Add labels to the k-points for a given axes"""
        # Label the kpoints
        labels = self.unfold.kpoint_labels
        kdist = self.unfold.get_kpoint_distances()

        # Explicit label indices
        tick_locs = []
        tick_labels = []
        for index, label in labels:
            if x_is_kidx:
                xloc = index
            else:
                xloc = kdist[index]
            ax.axvline(x=xloc, lw=0.5, color='k', ls=':', alpha=0.8)
            tick_locs.append(xloc)
            tick_labels.append(clean_latex_string(label))
        ax.set_xticks(tick_locs)
        ax.set_xticklabels(tick_labels)

    def plot_effective_mass(self,
                            eff: EffectiveMass,
                            engs: np.ndarray,
                            sf: np.ndarray,
                            eref: Union[None, float] = None,
                            save: Union[None, str] = None,
                            show: bool = False,
                            effective_mass_data: dict = None,
                            **kwargs):
        """
        Plot the effective masses on top of the spectral function.

        :param eff: An `EffectiveMass` object used for plotting.
        :param engs: The energies of the spectral functions.
        :param sf: An array of the spectral function.
        :param eref: Reference energy to be used - this energy will be set as zero.
        :param effective_mass_data: Calculated data of the effective masses.

        :returns: A figure with the data used for fitting effective mass plotted on top of the spectral function.
        """

        kcbm = eff.get_band_extrema(mode='cbm')[0]
        kvbm = eff.get_band_extrema(mode='vbm')[0]
        all_k = sorted(list(set(list(kcbm) + list(kvbm))))
        kdist = self.unfold.get_kpoint_distances()
        xwidth = 0.2
        if eref is None:
            eref = self.unfold.calculated_quantities['vbm']

        fig, axes = plt.subplots(1, len(all_k))
        if effective_mass_data is None:
            effective_mass_data = eff.get_effective_masses()

        # Plot the spectral function
        for (ik, ax) in zip(all_k, axes):
            self.plot_spectral_function(engs, sf, ax=ax, eref=eref, **kwargs)
            xk = kdist[ik]
            xlim = (xk - xwidth / 2, xk + xwidth / 2)
            ax.set_xlim(xlim)
            ax.set_title(f'Kpoint: {ik}')

        elec = effective_mass_data['electrons']
        # Plot the detected effective mass fitting data on top
        for entry in elec:
            ik = entry['kpoint_index']
            iax = all_k.index(ik)
            x = entry['raw_data']['kpoint_distances']
            y = entry['raw_data']['effective_energies']
            axes[iax].plot(x, np.asarray(y) - eref, '-o', color='C1')

        hole = effective_mass_data['holes']
        for entry in hole:
            ik = entry['kpoint_index']
            iax = all_k.index(ik)
            x = entry['raw_data']['kpoint_distances']
            y = entry['raw_data']['effective_energies']
            axes[iax].plot(x, np.asarray(y) - eref, '-o', color='C2')
        if save:
            fig.savefig(save)

        if show:
            fig.show()
        return fig

    def plot_spectral_weights(
        self,
        figsize=(4, 3),
        ylim=(-3, 3),
        dpi: float = 150,
        factor: float = 3.0,
        eref: Union[None, float] = None,
        color: str = 'C1',
        alpha: float = 0.5,
        save: Union[None, str] = None,
        ax: Union[None, plt.Axes] = None,
        show: bool = False,
        title: Union[None, str] = None,
    ):
        """
        Plot the spectral weights.

        ```{note}
        The reduction of symmetry means there can be multiple supercell kpoints for each
        primitive cell kpoint. When using this scattering plot representation, the markers can
        overlap each other leading to misrepresentations of the actual effective band structure.
        ```

        However, this function is still useful when: 1. the symmetry splitting is turned off. 2.
        direct visualization of the underlying spectral weight is needed. 3. Check the correctness
        of effective mass extraction.

        :param eref: Reference energy to be used - this energy will be set as zero.
        :param figsize: Size of the figure.
        :param ylim: Plotting limit for the y-axis, with respect to `eref`.
        :param dpi: DPI of the generated graph.
        :param alpha: Alpha for the markers.
        :param save: Name of the file where the generated figure is saved.
        :param ax: Existing plotting axes to be be used (list if having two spin channels).
        :param factor: Scaling factor for the marker size.
        :param color: Color of the markers.
        :param title: Title for the generated plot.

        :returns: A Figure with the spectral weights plotted as a scatter plot.
        """
        unfold = self.unfold
        kdist = unfold.get_kpoint_distances()
        # a list of spectral weights each with (nspin, nkset, nbands, 2)
        sws = unfold.get_spectral_weights()
        nspin = sws[0].shape[0]

        nb = sws[0].shape[2]

        if eref is None:
            eref = unfold.calculated_quantities.get('vbm', 0.0)

        if ax is None:
            if nspin == 1:
                fig, ax = plt.subplots(1, 1, figsize=figsize, dpi=dpi)
                axes = [ax]
            else:
                fig, axes = plt.subplots(1, 2, figsize=figsize, dpi=dpi)
        else:
            if not isinstance(ax, list):
                axes = [ax]
            else:
                axes = ax
            fig = axes[0].figure

        kweights = unfold.expansion_results['weights']
        xs = []
        ys = []
        ss = []
        # Each spin channel
        for (ispin, ax_) in zip(range(nspin), axes):
            # Collect data
            for (ik, sw) in enumerate(sws):
                # Each sub kpoint
                for isubk in range(sw.shape[1]):
                    x = np.repeat(kdist[ik], nb)
                    y = sw[ispin, isubk, :, 0] - eref
                    marker_size = (sw[ispin, isubk, :, 1] * factor) * kweights[ik][isubk]
                    xs.append(x)
                    ys.append(y)
                    ss.append(marker_size)

            ax_.scatter(
                np.vstack(xs),
                np.vstack(ys),
                s=np.vstack(ss),
                color=color,
                alpha=alpha,
                lw=0.0,
            )
            ax_.set_xlim(0, kdist.max())
            ax_.set_ylim(ylim)
            ax_.set_ylabel('Energy [eV]', labelpad=5)
            if title:
                ax_.set_title(title)
            self._add_kpoint_labels(ax_)

        fig.tight_layout(pad=0.2)

        if save:
            fig.savefig(save, dpi=300)
        if show:
            fig.show()
        return fig

    def plot_projected(
        self,
        procar: Union[str, list],
        eref=None,
        gamma=False,
        npoints=2000,
        sigma=0.2,
        ncl=False,
        symm_average=True,
        figsize=(4, 3),
        ylim=(-3, 3),
        dpi=150,
        vscale=1.0,
        contour_plot=False,
        alpha=1.0,
        save=False,
        ax=None,
        vmin=None,
        vmax=None,
        cmap='PuRd',
        show=False,
        title=None,
        atoms_idx=None,
        orbitals=None,
        intensity=1.0,
        use_subplot=False,
        colors=None,
        colorspace='lab',
    ):
        """
        Plot projected sepctral function onto multiple subplots or a single plot with colormapping.

        This simply computes the spectral function at each orbital/atoms sites and plot them onto
        multiple subplots. The columns are for each orbital and the rows are for each spin channel.

        :param procar: Name of names of the `PROCAR` files.

        :returns: Generated plot.
        """

        unfoldset = self.unfold
        unfoldset.load_procar(procar)
        nspins = unfoldset.calculated_quantities['spectral_weights_per_set'][0].shape[0]

        atoms_idx_subplots = atoms_idx.split('|')
        if orbitals is not None:
            orbitals_subsplots = orbitals.split('|')

            # Special case: if only one set is passed, apply it to all atomic specifications
            if len(orbitals_subsplots) == 1:
                orbitals_subsplots = orbitals_subsplots * len(atoms_idx_subplots)

            if len(orbitals_subsplots) != len(atoms_idx_subplots):
                raise ValueError('The number of elected orbitals and atoms indices are not matched.')
        # If not set, use all for all subsets
        else:
            orbitals_subsplots = ['all'] * len(atoms_idx_subplots)

        # Load the data

        nsub = len(atoms_idx_subplots)

        all_sf = []
        vmaxs = []

        if eref is None:
            eref = unfoldset.calculated_quantities.get('vbm', 0.0)

        # Collect spectral functions and scale
        for this_idx, this_orbitals in zip(atoms_idx_subplots, orbitals_subsplots):
            # Setup the atoms_idx and orbitals
            this_idx, this_orbitals = process_projection_options(this_idx, this_orbitals)
            eng, spectral_function = unfoldset.get_spectral_function(gamma=gamma,
                                                                     npoints=npoints,
                                                                     sigma=sigma,
                                                                     ncl=ncl,
                                                                     atoms_idx=this_idx,
                                                                     orbitals=this_orbitals,
                                                                     symm_average=symm_average)
            all_sf.append(spectral_function)

            emin, emax = ylim
            # Clip the effective range
            mask = (eng < (emax + eref)) & (eng > (emin + eref))
            vmin = spectral_function[:, mask, :].min()
            vmax = spectral_function[:, mask, :].max()
            vmax = (vmax - vmin) * vscale + vmin
            vmaxs.append(vmax)

        # Workout the vmax and vmin
        vmax = max(vmaxs)
        vmin = 0.

        if use_subplot:
            fig, axs = plt.subplots(nspins, len(atoms_idx_subplots), sharex=True, sharey=True, squeeze=False, figsize=(3.0 * nsub, 4.0))
            # Plot the spectral function with constant colour scales
            for spectral_function, i in zip(all_sf, range(len(atoms_idx_subplots))):
                plotter = UnfoldPlotter(unfoldset)
                plotter.plot_spectral_function(
                    eng,
                    spectral_function,
                    eref=eref,
                    save=save,
                    vscale=vscale,
                    vmax=vmax,
                    vmin=vmin,
                    cmap=cmap,
                    title=title,
                    ax=axs[:, i].tolist(),
                    figsize=figsize,
                    show=show,
                    contour_plot=contour_plot,
                    alpha=alpha,
                    dpi=dpi,
                    ylim=ylim,
                )
        else:
            # Make a combined plot
            sf_size = all_sf[0].shape
            stacked_sf = np.stack(all_sf, axis=-1).reshape(np.prod(sf_size), len(all_sf))

            # Construct the color basis
            if colors is None:
                default_colors = ['r', 'g', 'b', 'purple']
                colors = default_colors[0:len(all_sf)]
            # Compute spectral weight data with RGB reshape it back into the shape (nengs, nk, 3)
            sf_rgb = interpolate_colors(colors, stacked_sf, colorspace, normalize=True).reshape(sf_size + (3,))
            sf_sum = np.sum(all_sf, axis=0)[:, :, :, None]
            sf_rgba = np.concatenate([sf_rgb, sf_sum], axis=-1)

            if ax is None:
                fig, axs = plt.subplots(1, nspins, figsize=figsize, squeeze=True)

            self._plot_spectral_function_rgba(
                eng,
                sf_rgba,
                eref=eref,
                save=save,
                title=title,
                intensity=intensity,
                ax=axs,
                figsize=figsize,
                show=show,
                dpi=dpi,
                ylim=ylim,
            )

        return fig

    @staticmethod
    def plot_effective_mass_fit(efm: EffectiveMass,
                                npoints: int = 3,
                                carrier: str = 'electrons',
                                idx: int = 0,
                                ax: Union[plt.Axes, None] = None,
                                save: Union[None, str] = None,
                                dpi: float = 120):
        """
        Plot detected band edges and the fitted effective masses.

        :param efm: `EffectiveMass` object for plotting.
        :param npoints: The number of points to be used for fitting.
        :param carrier: Type of the charge carrier, e.g. `electrons` or `holes`.
        :param idx: Index for the detected effective mass of the same kind.
        :param ax: `plt.Axes` used for plotting.
        :param save: Name of the file used for saveing.
        :param dpi: DPI of the figure when saving.

        :return: A figure with plotted data.
        """
        data = efm.get_effective_masses(npoints=npoints)[carrier][idx]

        x = data['raw_data']['kpoint_distances']
        y = data['raw_data']['effective_energies']
        me = data['effective_mass']
        x1, y1 = fitted_band(x, me)
        if ax is None:
            fig, ax = plt.subplots(1, 1)
        ax.plot(x, y, 'x-', label='Energy ')
        ax.plot(x1, y1 + y[0], label='fitted')
        ax.legend()
        if save:
            fig.savefig(save, dpi=dpi)
        return fig


def interpolate_colors(colors: list, weights: list, colorspace='lab', normalize=True):
    """
    Interpolate colors at a number of points within a colorspace.

    :param colors: A list of colors specified in any way supported by matplotlib.
    :param weights: A list of weights with the shape (n, N). Where the N values of
      the last axis give the amount of N colors supplied in `colors`.
    :param colorspace: The colorspace in which to perform the interpolation. The
            allowed values are rgb, hsv, lab, luvlc, lablch, and xyz.

    :returns: A list of colors, specified in the rgb format as a (n, 3) array.
    """

    # Set up and check the color spaces
    colorspace_mapping = {
        'rgb': sRGBColor,
        'hsv': HSVColor,
        'lab': LabColor,
        'luvlch': LCHuvColor,
        'lablch': LCHabColor,
        'xyz': XYZColor,
    }

    if colorspace not in list(colorspace_mapping.keys()):
        raise ValueError(f'colorspace must be one of {colorspace_mapping.keys()}')

    colorspace = colorspace_mapping[colorspace]

    # Convert matplotlib color specification to colormath sRGB
    colors_srgb = [sRGBColor(*to_rgb(c)) for c in colors]

    colors_basis = [np.array(convert_color(srgb, colorspace, target_illuminant='d50').get_value_tuple()) for srgb in colors_srgb]

    # ensure weights is a numpy array
    weights = np.asarray(weights)

    # Normalise the weights if needed
    if normalize:
        weights = weights / np.linalg.norm(weights, axis=1)[:, None]

    # perform the interpolation in the colorspace basis
    interpolated_colors = colors_basis[0] * weights[:, 0][:, None]
    for i in range(1, len(colors_basis)):
        interpolated_colors += colors_basis[i] * weights[:, i][:, None]

    # convert the interpolated colors back to RGB
    rgb_colors = [convert_color(colorspace(*c), sRGBColor).get_value_tuple() for c in interpolated_colors]
    rgb_colors = np.stack(rgb_colors, axis=0)

    # ensure all rgb values are less than 1 (sometimes issues in interpolation gives
    np.clip(rgb_colors, 0, 1, rgb_colors)
    return rgb_colors



"""
Physical constants used in VASP
"""

__all__ = [
    'AUTOA', 'RYTOEV', 'CLIGHT', 'EVTOJ', 'AMTOKG', 'BOLKEV', 'BOLK', 'EVTOKCAL', 'PI', 'TPI', 'CITPI', 'FELECT', 'EDEPS', 'HSQDTM',
    'MAGMOMTOENERGY', 'MOMTOMOM', 'AUTOA2', 'AUTOA3', 'AUTOA4', 'AUTOA5', 'AUTDEBYE'
]

#  Some important Parameters, to convert to a.u.
#  - AUTOA     =  1. a.u. in Angstroem
#  - RYTOEV    =  1 Ry in Ev
#  - EVTOJ     =  1 eV in Joule
#  - AMTOKG    =  1 atomic mass unit ("proton mass") in kg
#  - BOLKEV    =  Boltzmanns constant in eV/K
#  - BOLK      =  Boltzmanns constant in Joule/K

AUTOA = 0.529177249
RYTOEV = 13.605826
CLIGHT = 137.037  # speed of light in a.u.
EVTOJ = 1.60217733E-19
AMTOKG = 1.6605402E-27
BOLKEV = 8.6173857E-5
BOLK = BOLKEV * EVTOJ
EVTOKCAL = 23.06

# FELECT    =  (the electronic charge)/(4*pi*the permittivity of free space)
#         in atomic units this is just e^2
# EDEPS    =  electron charge divided by the permittivity of free space
#         in atomic units this is just 4 pi e^2
# HSQDTM    =  (plancks CONSTANT/(2*PI))**2/(2*ELECTRON MASS)
#
PI = 3.141592653589793238
TPI = 2 * PI
CITPI = 1j * TPI
FELECT = 2 * AUTOA * RYTOEV
EDEPS = 4 * PI * 2 * RYTOEV * AUTOA
HSQDTM = RYTOEV * AUTOA * AUTOA

# vector field A times momentum times e/ (2 m_e c) is an energy
# magnetic moments are supplied in Bohr magnetons
# e / (2 m_e c) A(r) p(r)    =  energy
# e / (2 m_e c) m_s x ( r - r_s) / (r-r_s)^3 hbar nabla    =
# e^2 hbar^2 / (2 m_e^2 c^2) 1/ lenght^3    =  energy
# conversion factor from magnetic moment to energy
# checked independently in SI by Gilles de Wijs

MAGMOMTOENERGY = 1 / CLIGHT**2 * AUTOA**3 * RYTOEV

# dimensionless number connecting input and output magnetic moments
# AUTOA e^2 (2 m_e c^2)
MOMTOMOM = AUTOA / CLIGHT / CLIGHT / 2
AUTOA2 = AUTOA * AUTOA
AUTOA3 = AUTOA2 * AUTOA
AUTOA4 = AUTOA2 * AUTOA2
AUTOA5 = AUTOA3 * AUTOA2

# dipole moment in atomic units to Debye
AUTDEBYE = 2.541746



# -*- coding: utf-8 -*-
"""
The main module for unfolding workflow and algorithm
"""
# pylint: disable=invalid-name,protected-access,too-many-locals

############################################################
import re
from typing import Union, List, Tuple
from pathlib import Path

import numpy as np
import matplotlib.pyplot as plt
from matplotlib.colors import ListedColormap, hex2color
from monty.json import MSONable
from monty.serialization import loadfn
from tqdm import tqdm
import ase
import spglib

from easyunfold import __version__
from .wavecar import Wavecar
from .procar import Procar

############################################################


def get_symmetry_dataset(atoms: ase.Atoms, symprec: float = 1e-5):
    """Get the symmetry dataset using spglib"""
    return spglib.get_symmetry_dataset((atoms.get_cell(), atoms.get_scaled_positions(), atoms.get_atomic_numbers()), symprec=symprec)


def find_K_from_k(k: np.ndarray, M: np.ndarray):
    """
    Get the K vector of the supercell onto which the k vector of the primitive
    cell folds. The unfolding vector G, which satisfy the following equation,
    is also returned.

    $$
        \\vec{k} = \\vec{K} + \\vec{G}
    $$

    where G is a reciprocal space vector of supercell
    """

    M = np.array(M)
    Kc = np.dot(k, M.T)
    G = np.array(np.round(Kc), dtype=int)
    KG = Kc - np.round(Kc)

    return KG, G


def rotate_kpt(k: np.ndarray, opt: np.ndarray):
    """
    Apply rotation to a kpoint based on the rotations of the crystals (in the real space)

    NOTE: The rotation matrix should be the one that act on fractional coordinates, e.g. from spglib.
    """
    rot = k @ opt
    return rot - np.rint(rot)


def expand_K_by_symmetry(kpt: Union[list, np.ndarray], opts_pc: np.ndarray, opts_sc: np.ndarray, time_reversal: bool = True):
    """
    Expend the sampling of the PC kpoints due to symmetry breaking of the supercell cell.

    :returns: Expanded kpoints and corresponding weights for each primitive cell kpoint.
    """
    kpt_orig = np.asarray(kpt)

    # Find distinct images of the kpoints in the PC
    pc_distinct = [kpt_orig]
    for opt in opts_pc:
        k_equiv = rotate_kpt(kpt_orig, opt)
        k_equiv_neg = k_equiv * -1
        found = False
        for k_ in pc_distinct:
            if np.allclose(k_equiv, k_):
                found = True
            if time_reversal and np.allclose(k_equiv_neg, k_):
                found = True
        if not found:
            pc_distinct.append(k_equiv)
    weights = np.ones(len(pc_distinct))
    # Now each for uniqueness in the supercell
    for i, kpt_ in enumerate(pc_distinct):
        if weights[i] == 0:
            continue

        for opt in opts_sc:
            k_equiv = rotate_kpt(kpt_, opt)
            k_equiv_neg = k_equiv * -1
            for j, k_ in enumerate(pc_distinct):
                # Skip if it is the same point, or the other point has been taken
                if i == j or weights[j] == 0:
                    continue
                if np.allclose(k_equiv, k_) or (time_reversal and np.allclose(k_equiv_neg, k_)):
                    weights[i] += weights[j]
                    weights[j] = 0

    out_points = []
    out_weights = []
    for i, kpt_ in enumerate(pc_distinct):
        if weights[i] != 0:
            out_points.append(kpt_)
            out_weights.append(weights[i])
    assert sum(out_weights) == len(pc_distinct)
    # The first kpoint of the set should always be the original kpoint
    assert out_points[0] is kpt_orig
    out_weights = np.array(out_weights) / sum(out_weights)
    return out_points, out_weights


class UnfoldKSet(MSONable):
    """
    High-level interface for unfolding and serialization.
    """

    _VERSION = '0.1.0'

    def __init__(self,
                 M: np.ndarray,
                 kpts_pc: list,
                 pc_latt: np.ndarray,
                 pc_opts: np.ndarray,
                 sc_opts: np.ndarray,
                 time_reversal: bool = True,
                 expand: bool = True,
                 metadata: Union[None, dict] = None,
                 expansion_results: Union[None, dict] = None,
                 calculated_quantities: Union[None, dict] = None,
                 kpoint_labels: Union[None, list] = None):
        """
        Instantiate an `UnfoldKSet` object.

        :param M: The supercell transformation matrix
        :param kpts_pc: A list of kpoints in the PC
        :param pc_latt: A 3x3 matrix of row lattice vectors of the primitive cell
        :param pc_opts: Symmetry operations of the primitive cell
        :param sc_opts: Symmetry operations of the supercell
        :param expand: Whether to expand the kpoint to take account of broken symmetry or not
        :param expansion_results: Using existing results of symmetry expansion
        :param calculated_quantities: Existing calculated quantities
        :param kpoint_labels: Labels of the kpoints as passed in `pc_latt` as a list of tuples `(<idx>, <label>)`
        """
        # Basic properties - needed to recreate the object
        self.kpts_pc = kpts_pc
        self.pc_latt = pc_latt
        self.pc_opts = pc_opts
        self.sc_opts = sc_opts
        self.expand = expand
        self.M = M
        self.expansion_results = expansion_results
        self.time_reversal = time_reversal
        self.calculated_quantities = {} if not calculated_quantities else calculated_quantities
        self.kpoint_labels = kpoint_labels
        if metadata is None:
            metadata = {}
        self.metadata = metadata
        self.metadata['program_version'] = __version__
        self.transient_quantities = {}

        # Transient properties
        self.reduced_sckpts = None
        self.reduced_sckpts_map = None
        if self.expansion_results is None:
            self.expand_pc_kpoints()

    @property
    def is_calculated(self) -> bool:
        """Show the status of the work"""
        return bool(self.calculated_quantities)

    @property
    def has_averaged_spectral_weights(self) -> bool:
        """Return True if the spectral weights stored is averaged"""
        return self.calculated_quantities.get('spectral_weights_is_averaged', False)

    @classmethod
    def from_atoms(cls,
                   M: np.ndarray,
                   kpts_pc: list,
                   pc: ase.Atoms,
                   sc: ase.Atoms,
                   time_reversal: bool = True,
                   expand=True,
                   symprec: float = 1e-5):
        """
        Initialise from primitive cell and supercell atoms

        :param M: The supercell transformation matrix
        :param kpts_pc: A list of kpoints in the PC
        :param pc: Primitive cell structure
        :param sc: Supercell structure
        :param time_reversal: Assumes time-reversal symmetry
        :param expand: Whether to expand the kpoint to take account of broken symmetry or not
        :param symprec: Symmetry detection precision

        """
        pc_symm_data = get_symmetry_dataset(pc, symprec=symprec)
        sc_symm_data = get_symmetry_dataset(sc, symprec=symprec)
        return cls(
            M=M,
            kpts_pc=kpts_pc,
            pc_latt=np.asarray(pc.cell),
            pc_opts=pc_symm_data['rotations'],
            sc_opts=sc_symm_data['rotations'],
            time_reversal=time_reversal,
            expand=expand,
            metadata={
                'symmetry_dataset_pc': pc_symm_data,
                'symmetry_dataset_sc': sc_symm_data,
            },
        )

    @classmethod
    def from_file(cls, fname: str):
        """Load from a file"""
        return loadfn(fname)

    def expand_pc_kpoints(self) -> None:
        """Compute the pc kpoints to be unfolded into"""
        expended_k = []
        expended_weights = []
        for kpt in self.kpts_pc:
            # Record the expanded kpoints and corresponding weights
            if self.expand:
                kset, weights = expand_K_by_symmetry(kpt, self.pc_opts, self.sc_opts, time_reversal=self.time_reversal)
            else:
                # Just take the original point and set the weight to be unity
                kset = [kpt]
                weights = np.array([1.0])
            expended_k.append(kset)
            expended_weights.append(weights)
        # Save as the attribute
        self.expansion_results = {'kpoints': expended_k, 'weights': expended_weights}

    def __repr__(self) -> str:
        return f'<UnfoldKSet with {self.nkpts_expand}/{self.nkpts_orig} kpoints based on {self.nsymm_expand}/{self.nsymm_orig} symm ops'

    @property
    def nsymm_orig(self) -> int:
        """Number of symmetry operation in the original cell"""
        return self.pc_opts.shape[0]

    @property
    def nsymm_expand(self) -> int:
        """Number of symmetry operation in the original cell"""
        return self.sc_opts.shape[0]

    @property
    def nkpts_orig(self) -> int:
        """Total number of unexpanded kpoints"""
        return len(self.expansion_results['kpoints'])

    @property
    def nkpts_expand(self) -> int:
        """Total number of expanded kpoints"""
        return sum(map(len, self.expansion_results['kpoints']))

    def generate_sc_kpoints(self) -> None:
        """
        Generate the supercell kpoints to be calculated.
        Results are stored into `self.expansion_results`.
        """

        assert bool(self.expansion_results)
        expended_sc = []
        all_sc = []
        # Find the SC kpoint for each PC kpoint
        for kset in self.expansion_results['kpoints']:
            this_k = []
            for kpt in kset:
                sc_k, _ = find_K_from_k(kpt, self.M)
                this_k.append(sc_k)
            expended_sc.append(this_k)
            all_sc.extend(this_k)
        # We now have bunch of supercell kpoints for each set of expanded kpoints
        # Try to find duplicated SC kpoints
        reduced_sckpts, sc_kpts_map = remote_duplicated_kpoints(all_sc, return_map=True)
        sc_kpts_map = list(sc_kpts_map)

        # Mapping between the pckpts to the redcued sckpts
        reduced_sc_map = []
        for sc_set in expended_sc:
            map_indx = []
            for _ in sc_set:
                map_indx.append(sc_kpts_map.pop(0))
            reduced_sc_map.append(map_indx)

        self.expansion_results['reduced_sckpts'] = reduced_sckpts
        self.expansion_results['reduced_sckpts_map'] = reduced_sc_map
        # A nested list that stores the indices of the sc kpts in the reduced_sckpts list

    def write_sc_kpoints(self, file: str, nk_per_split: Union[None, list] = None, scf_kpoints_and_weights: Union[None, list] = None):
        """
        Write the supercell kpoints to a file.

        :param file: Name of the file
        :param nk_per_split: Number of kpoints per split along the path
        :param scf_kpoints_and_weights: SCF kpoint and their weights needed for split-path calculations
        """
        if self.expansion_results.get('reduced_sckpts') is None:
            self.generate_sc_kpoints()
        kpoints = np.asarray(self.expansion_results['reduced_sckpts'])
        weights = None
        if nk_per_split is None:
            if scf_kpoints_and_weights:
                # Prepend with SCF kpoints
                kpoints, weights = concatenate_scf_kpoints(scf_kpoints_and_weights[0], scf_kpoints_and_weights[1], kpoints)

            write_kpoints(kpoints, file, comment='supercell kpoints', weights=weights)
        else:
            splits = [kpoints[i:i + nk_per_split] for i in range(0, kpoints.shape[0], nk_per_split)]
            for i_spilt, kpt in enumerate(splits):
                if scf_kpoints_and_weights:
                    kpt, weights = concatenate_scf_kpoints(scf_kpoints_and_weights[0], scf_kpoints_and_weights[1], kpt)
                write_kpoints(kpt, f'{file}_{i_spilt + 1:03d}', f'supercell kpoints split {i_spilt + 1}', weights)

    def write_pc_kpoints(self, file: str, expanded: bool = False):
        """Write the primitive cell kpoints"""
        if expanded:
            all_pc = []
            for tmp in self.expansion_results['kpoints']:
                all_pc.extend(tmp)
            write_kpoints(all_pc, file, comment='expanded primitive cell kpoints')
        else:
            write_kpoints(self.kpts_pc, file, comment='primitive cell kpoints')

    def _read_weights(self, wavecar: Union[str, List[str]], gamma: bool, ncl: bool, gamma_half: str):
        """
        Read the weights from the wavecar for all kpoints

        Returns the averaged weights and the original weights per set of kpoints
        """
        weights_per_set = []
        averaged_weights = []
        if not isinstance(wavecar, (list, tuple)):
            wavecar = [wavecar]
        # Load the WAVECAR unfold objects
        unfold_objs = [Unfold(self.M, name, gamma=gamma, lsorbit=ncl, gamma_half=gamma_half) for name in wavecar]
        # Record the VBM and the CBM values
        varray = np.array([obj.get_vbm_cbm() for obj in unfold_objs])
        self.calculated_quantities['vbm'] = float(varray[:, 0].max())
        self.calculated_quantities['cbm'] = float(varray[:, 1].min())
        # Read the spectral weights for each set of expanded kpoints
        for kset, weights in tqdm(zip(self.expansion_results['kpoints'], self.expansion_results['weights']),
                                  desc='kpt',
                                  total=len(self.expansion_results['kpoints'])):
            # Raw spectral weights
            sw = spectral_weight_multiple_source(kset, unfold_objs, self.M)
            weights_per_set.append(sw.copy())
            # Take weighted average
            for ik, w in enumerate(weights):
                sw[:, ik, :, :] *= w
            averaged_weights.append(sw.sum(axis=1))

        # Recreate the full weights array
        averaged_weights = np.stack(averaged_weights, axis=1)
        self.calculated_quantities['spectral_weights_per_set'] = weights_per_set
        self.calculated_quantities['version'] = self._VERSION

        return averaged_weights, weights_per_set

    def load_procar(self, procar: Union[str, List[str]], force: bool = False):
        """Read in PROCAR for band-based projection"""
        if self.procars and not force:
            pass

        if not isinstance(procar, (tuple, list)):
            procar = [procar]

        # Load the procars
        self.transient_quantities['procars'] = [Procar(path) for path in procar]
        # Construct mapping from the primitive cell kpoints to those in the PROCAR
        self.transient_quantities['procars_kmap'] = self._construct_procar_kmap()

    def _construct_procar_kmap(self) -> list:
        """Construct mapping from the set of kpoints to that in the PROCAR"""
        ksets = self.expansion_results['kpoints']
        kidx_procar_sets = []
        for kset in ksets:
            kidx_procar_sets.append([])
            for kpoint in kset:
                # Find the supercell kpoint
                K_super, _ = find_K_from_k(kpoint, self.M)
                # Search for kpoints in the procar
                found = False
                for iprocar, procar in enumerate(self.procars):
                    for ikpt, kprocar in enumerate(procar.kvecs[0]):
                        if np.allclose(K_super, kprocar):
                            kidx_procar_sets[-1].append([iprocar, ikpt])
                            found = True
                            break
                    if found:
                        break
                if found is False:
                    raise ValueError(f'Cannot found kpoint {K_super} in PROCAR files')
        return kidx_procar_sets

    @property
    def procars(self) -> Union[None, Procar]:
        """Loaded PROCARS"""
        return self.transient_quantities.get('procars')

    @property
    def procar_kmaps(self):
        """Loaded PROCAR kpoint mapping"""
        return self.transient_quantities.get('procars_kmap')

    def _get_spectral_weights(self,
                              wavecar,
                              npoints=2000,
                              sigma=0.01,
                              emin=None,
                              emax=None,
                              gamma=False,
                              ncl=False,
                              gamma_half='x',
                              also_spectral_function=False,
                              atoms_idx=None,
                              orbitals=None,
                              symm_average=True):
        """
        Fetch spectral weights from a wavecar and compute spectral function is requested

        Args:

            atomic_projects (tuple): A tuple of atoms and orbitals whose projected weights should be used.
        """
        # If WAVECAR is given - reload from the data
        if wavecar:
            self._read_weights(wavecar, gamma=gamma, ncl=ncl, gamma_half=gamma_half)
        else:
            if not self.is_calculated:
                raise RuntimeWarning('The spectral weights need to be calculated first - please pass the WAVECAR file(s).')

        # Use existing results
        if symm_average:
            sws = self.calculated_quantities['spectral_weights_per_set']
            kweight_sets = self.expansion_results['weights']
        else:
            # No averaging - we just return the first item of each set, which is the weight of the original set
            sws = [item[:, :1, :, :] for item in self.calculated_quantities['spectral_weights_per_set']]
            # Recreate the full weights array
            kweight_sets = [[1.0] for i in range(len(sws))]

        if also_spectral_function:
            if atoms_idx is not None:
                # Read in the project weights
                band_weight_sets = self.get_band_weight_sets(atoms_idx, orbitals)
            else:
                band_weight_sets = None
            e0, spectral_function = spectral_function_from_weight_sets(sws,
                                                                       kweight_sets,
                                                                       nedos=npoints,
                                                                       sigma=sigma,
                                                                       emin=emin,
                                                                       emax=emax,
                                                                       band_weight_sets=band_weight_sets)
            self.calculated_quantities['e0'] = e0
            self.calculated_quantities['spectral_function'] = spectral_function
            return sws, e0, spectral_function
        return sws

    def get_band_weight_sets(self,
                             atoms_idx: List[int],
                             orbitals: List[Union[List[str], str]],
                             procars: Union[None, List[str], str] = None) -> list:
        """
        Get weights array sets for bands

        Construct the weights of each band in same format of the kpoint set.
        Each item is an numpy array of (nspins, nbands), containing the summed weights over
        the passed atom indices and orbitals.

        :param atoms_idx: Indicies of the atoms to be selected
        :param orbitals: Orbitals to be selected for each atom
        :param procars: Names of the PROCAR files to be loaded

        :returns: A list of weights for each band at each expanded kpoint
        """
        if procars:
            self.load_procar(procars)
        if self.procars is None:
            raise RuntimeError('PROCAR files needs to be loaded')

        projs = [procar.get_projection(atoms_idx, orbitals) for procar in self.procars]
        # Construct band weighting, same structure as o
        band_weight_sets = []
        for kset in self.transient_quantities['procars_kmap']:
            band_weight_sets.append([])
            # Search
            for iprocar, kidx in kset:
                band_weight = projs[iprocar][:, kidx]
                band_weight_sets[-1].append(band_weight)
        return band_weight_sets

    def get_spectral_function(self,
                              wavecar: Union[None, Wavecar] = None,
                              npoints: int = 2000,
                              sigma: float = 0.1,
                              gamma: bool = False,
                              ncl: bool = False,
                              gamma_half: str = 'x',
                              symm_average: bool = True,
                              atoms_idx: Union[None, List[int]] = None,
                              orbitals: Union[None, str, List[str]] = None):
        """
        Compute and return the spectral function

        :param wavecar: The WAVECAR files to be used
        :param npoints: Number of points for the energy axis
        :param sigma: Smearing width for the Gaussian smearing
        :param gamma: Need to be set to `True` for Gamma-only calculation
        :param ncl: Need to be set to `True` for non-collinear magnetism calculation
        :param gamma_half: Flag used for reading WAVECAR
        :param symm_average: Whether to perform symmetry averaging
        :param atoms_idx: Indices for the atoms for projection
        :param orbitals: Indices for the atoms for projection

        :returns: A tuple of the energies and the spectral functioin.
        """
        _, e0, spectral_function = self._get_spectral_weights(wavecar,
                                                              npoints=npoints,
                                                              sigma=sigma,
                                                              gamma=gamma,
                                                              ncl=ncl,
                                                              gamma_half=gamma_half,
                                                              also_spectral_function=True,
                                                              atoms_idx=atoms_idx,
                                                              orbitals=orbitals,
                                                              symm_average=symm_average)
        return e0, spectral_function

    def get_spectral_weights(self, wavecar=None, gamma: bool = False, ncl: bool = False, gamma_half: str = 'x', symm_average: bool = True):
        """
        Return the spectral weights calculated

        :param wavecar: The WAVECAR file(s) for calculation. Not need if the weights has been calculated.
        :param gamma: Whether the WAVECAR files are from $\\Gamma$-only calculation.

        :returns: An array storing the spectral weights.

        """
        return self._get_spectral_weights(wavecar=wavecar,
                                          gamma=gamma,
                                          ncl=ncl,
                                          gamma_half=gamma_half,
                                          also_spectral_function=False,
                                          symm_average=symm_average)

    def as_dict(self) -> dict:
        """Convert the object into a dictionary representation"""
        output = {'@module': self.__class__.__module__, '@class': self.__class__.__name__, '@version': __version__}
        for key in [
                'M', 'kpts_pc', 'pc_latt', 'pc_opts', 'sc_opts', 'expansion_results', 'time_reversal', 'calculated_quantities',
                'kpoint_labels', 'expand', 'metadata'
        ]:
            output[key] = getattr(self, key)
        return output

    def get_kpoint_distances(self):
        """
        Distances between the kpoints along the path in the reciprocal space.
        This does not take account of the breaking of the path.
        :::{note}
        The reciprocal lattice vectors includes the $2\\pi$ factor, e.g. `np.linalg.inv(L).T * 2 * np.pi`.
        :::
        """
        kpts = self.kpts_pc
        pc_latt = self.pc_latt
        kpts_path = kpts @ np.linalg.inv(pc_latt).T * np.pi * 2  # Kpoint path in the reciprocal space
        dists = np.cumsum(np.linalg.norm(np.diff(kpts_path, axis=0), axis=-1))
        dists = np.append([0], dists)

        return dists


def LorentzSmearing(x, x0, sigma=0.02):
    r"""
    Simulate the Delta function by a Lorentzian shape function
    $$
        \Delta(x) = \lim_{\sigma\to 0}  Lorentzian
    $$
    """

    return 1. / np.pi * sigma**2 / ((x - x0)**2 + sigma**2)


def GaussianSmearing(x, x0, sigma=0.02):
    r"""
    Simulate the Delta function by a Lorentzian shape function
    $$
        \Delta(x) = \lim_{\sigma\to 0} Gaussian
    $$
    """

    return 1. / (np.sqrt(2 * np.pi) * sigma) * np.exp(-(x - x0)**2 / (2 * sigma**2))


def remote_duplicated_kpoints(kpoints: list, return_map=False, decimals=6):
    """
    remove duplicate kpoints in the list.
    """
    kpoints = np.asarray(kpoints)
    _, kid, inv_kid = np.unique(
        np.round(kpoints, decimals),
        axis=0,
        return_index=True,
        return_inverse=True,
    )
    reducedK = kpoints[kid]

    if return_map:
        return reducedK, inv_kid
    return reducedK


def write_kpoints(kpoints: Union[np.ndarray, list], outpath: str = 'KPOINTS', comment: str = '', weights: Union[None, List[float]] = None):
    """
    Write kpoints to VASP KPOINTS file

    :param kpoints: A list of kpoints to be written
    :param outpath: Path to the output file
    :param comments: Comments to be put into the `KPOINTS` file
    :param weights: If given, the weighting of the kpoints, otherwise all kpoints are equal weighted.
    """
    kpoints = np.asarray(kpoints)
    nkpts = kpoints.shape[0]

    with open(outpath, 'w', encoding='utf-8') as vaspkpt:
        vaspkpt.write(comment + '\n')
        vaspkpt.write(f'{nkpts}\n')
        vaspkpt.write('Rec\n')
        for ik in range(nkpts):
            if weights is not None:
                line = f'  {kpoints[ik, 0]:12.8f} {kpoints[ik, 1]:12.8f} {kpoints[ik,2]:12.8f}  {weights[ik]:12.8f}\n'
            else:
                line = f'  {kpoints[ik, 0]:12.8f} {kpoints[ik, 1]:12.8f} {kpoints[ik,2]:12.8f}  1.0\n'
            vaspkpt.write(line)


def read_kpoints(path='KPOINTS'):
    """
    Read kpoints from a KPOINTS file containing reciprocal space coordinates (fractional)

    :returns: The kpoints, the comment, the labels, and the weights at each kpoint.
    """
    content = Path(path).read_text(encoding='utf-8').split('\n')
    comment = content[0]
    nkpts = int(content[1])
    if content[2].lower().startswith('lin'):
        return read_kpoints_line(content)
    assert content[2].lower().startswith('rec'), 'Only Reciprocal space KPOINT file is supported'
    kpts = []
    labels = []
    weights = []
    ik = 0
    for line in content[3:]:
        tokens = line.split()
        this_kpt = [float(value) for value in tokens[:3]]
        weights.append(float(tokens[3]))

        if len(tokens) >= 5:
            labels.append([ik, tokens[4]])

        kpts.append(this_kpt)
        ik += 1
        if ik == nkpts:
            break
    return kpts, comment, labels, weights


def read_kpoints_line(content, density=20):
    """
    Read kpoints in the VASP KPOINTS file line mode

    Resolve to explicit kpoints

    :returns: The kpoints, the comment, the labels, and the weights at each kpoint
    """
    comment = content[0]
    density = int(content[1]) if content[1] else density

    assert content[2].lower().startswith('lin'), 'Only Line mode KPOINT file is supported'
    assert content[3].lower().startswith('rec'), 'Only Reciprocal coorindates are supported!'

    segs = []
    labels = []
    for line in content[4:]:
        tokens = line.split()
        if not tokens:
            continue
        point = [float(x) for x in tokens[:3]]
        labels.append(tokens[-1])
        segs.append(point)
    # Process the segments
    kpoints = []
    labels_loc = []
    for i in range(int(len(segs) / 2)):
        k1 = segs[i * 2]
        k2 = segs[i * 2 + 1]
        # Check for duplicate end point
        if kpoints and kpoints[-1] == k1:
            kpoints.pop()
            labels_loc.pop()
        this_seg = np.linspace(k1, k2, density)
        labels_loc.append((len(kpoints), labels[i]))
        kpoints.extend(this_seg.tolist())
        labels_loc.append((len(kpoints) - 1, labels[i + 1]))
    return kpoints, comment, labels_loc, None


def make_kpath(kbound: List[float], nseg=40):
    """
    Return a list of kpoints defining the path between the given kpoints.
    """
    kbound = np.array(kbound, dtype=float)
    kdist = np.diff(kbound, axis=0)

    kpath = [kbound[ii] + kdist[ii] / nseg * nk for ii in range(len(kdist)) for nk in range(nseg)]
    kpath.append(kbound[-1])
    return kpath


def clean_latex_string(label: str):
    """
    Clean up latex labels and convert if necessary

    :returns: Cleaned tag string
    """
    if label == 'G':
        label = r'$\mathrm{\mathsf{\Gamma}}$'
    elif label.startswith('\\'):  ## This is a latex formatted label already
        label = f'$\\mathrm{{\\mathsf{{{label}}}}}$'
    else:
        label = r'$\mathrm{\mathsf{' + label + r'}}$'
    return label


def spectral_function_from_weight_sets(spectral_weight_sets: np.ndarray,
                                       kweight_sets: list,
                                       nedos: int = 4000,
                                       sigma: float = 0.02,
                                       emin=None,
                                       emax=None,
                                       band_weight_sets=None):
    r"""
    Generate the spectral function

    $$
        A(k_i, E) = \sum_m P_{Km}(k_i)\Delta(E - Em)
    $$

    Where the \Delta function can be approximated by Lorentzian or Gaussian
    function.

    :param band_weight_sets: Additional weighting for each band, used for generating
          projection onto atomic orbitals.
    """

    nk = len(spectral_weight_sets)
    ns = spectral_weight_sets[0].shape[0]
    spectral_function = np.zeros((ns, nedos, nk), dtype=float)

    emin = spectral_weight_sets[0][:, :, :, 0].min() if emin is None else emin
    emax = spectral_weight_sets[0][:, :, :, 0].max() if emax is None else emax
    e0 = np.linspace(emin - 5 * sigma, emax + 5 * sigma, nedos)

    for ispin in range(ns):
        for ii in range(nk):  # Iterate through kpoint sets (of primitive cell kpoints)
            for jj in range(spectral_weight_sets[ii].shape[1]):
                kweight = kweight_sets[ii][jj]
                E_Km = spectral_weight_sets[ii][ispin, jj, :, 0]
                P_Km = spectral_weight_sets[ii][ispin, jj, :, 1]
                if band_weight_sets is not None:
                    P_Km = P_Km * band_weight_sets[ii][jj][ispin, :P_Km.shape[0]]
                # Take weighted average spectral functions
                spectral_function[ispin, :,
                                  ii] += np.sum(LorentzSmearing(e0[:, np.newaxis], E_Km[np.newaxis, :], sigma=sigma) * P_Km[np.newaxis, :],
                                                axis=1) * kweight
    return e0, spectral_function


############################################################


class Unfold():
    """
    Low lever interface for performing unfolding related calculations.
    obtain the effective band structure (EBS).

    :::{admonition} Reference
    "Extracting E versus k effective band structure from supercell
     calculations on alloys and impurities"
    Phys. Rev. B 85, 085201 (2012)
    :::
    """

    def __init__(self, M=None, wavecar: str = 'WAVECAR', gamma: bool = False, lsorbit: bool = False, gamma_half: str = 'x', verbose=False):
        """
        Initialization.

        M is the transformation matrix between supercell and primitive cell:

        ```python
        M = np.dot(A, np.linalg.inv(a))
        ```

        In real space, the basis vectors of Supercell (A) and those of the
        primitive cell (a) satisfy:

        ```python
        A = np.dot(M, a);      a = np.dot(np.linalg.inv(M), A)
        ```

        Whereas in reciprocal space

        ```python
        b = np.dot(M.T, B);    B = np.dot(np.linalg.inv(M).T, b)
        ```

        wavecar is the location of VASP WAVECAR file that contains the
        wavefunction information of a supercell calculation.
        """

        # Whether the WAVECAR is a gamma-only version
        self._lgam = gamma
        self._lsoc = lsorbit

        self.M = np.array(M, dtype=float)
        assert self.M.shape == (3, 3), 'Shape of the tranformation matrix must be (3,3)'

        self.wfc = Wavecar(wavecar, lsorbit=self._lsoc, lgamma=self._lgam, gamma_half=gamma_half)
        # all the K-point vectors
        self.kvecs = self.wfc._kvecs
        # all the KS energies in shape (ns, nk, nb)
        self.bands = self.wfc._bands

        # spectral weight for all the kpoints
        self.SW = None

        self.verbose = verbose

    def get_vbm_cbm(self, thresh: float = 1e-8) -> Tuple[float, float]:
        """Locate the VBM from the WAVECAR"""
        occ = self.wfc._occs

        occupied = np.abs(occ) > thresh
        vbm = float(self.bands[occupied].max())
        cbm = float(self.bands[~occupied].min())
        return vbm, cbm

    def get_ovlap_G(self, ikpt: int = 1, epsilon: float = 1E-5) -> Tuple[np.ndarray, np.ndarray]:
        """
        Get subset of the reciprocal space vectors of the supercell,
        specifically the ones that match the reciprocal space vectors of the
        primitive cell.
        """

        assert 1 <= ikpt <= self.wfc._nkpts, 'Invalid K-point index!'

        # Reciprocal space vectors of the supercell in fractional unit
        Gvecs = self.wfc.get_gvectors(ikpt=ikpt)

        if self._lgam:
            nplw = Gvecs.shape[0]
            tmp = np.zeros((nplw * 2 - 1, 3), dtype=int)
            # the gvectors of Gamma version only contains half the number of a
            # normal version.
            tmp[:nplw, ...] = Gvecs
            tmp[nplw:, ...] = -Gvecs[1:, ...]  # G' = -G

            Gvecs = tmp

        # Shape of Gvecs: (nplws, 3)

        # Reciprocal space vectors of the primitive cell
        gvecs = np.dot(Gvecs, np.linalg.inv(self.M).T)
        # Deviation from the perfect sites
        gd = gvecs - np.round(gvecs)
        match = np.alltrue(np.abs(gd) < epsilon, axis=1)

        return Gvecs[match], Gvecs

    def find_K_index(self, K0: np.ndarray) -> int:
        """
        Find the index of a point `K0`.
        """

        for ii in range(self.wfc._nkpts):
            if np.alltrue(np.abs(self.wfc._kvecs[ii] - K0) < 1E-5):
                return ii + 1
        raise ValueError('Cannot find the corresponding K-points in WAVECAR!')

    def k2K_map(self, kpath) -> List[int]:
        """
        Return the map from primitive-cell k-points to supercell k-points.
        """

        return [self.find_K_index(find_K_from_k(k, self.M)[0]) - 1 for k in kpath]

    def spectral_weight_k(self, k0, whichspin=1):
        r"""
        Spectral weight for a given $k$:

        $$
            P_{Km}(k) = \sum_n |<Km | kn>|^2
        $$

        which is equivalent to
        $$
            P_{Km}(k) = \sum_{G} |C_{Km}(G + k - K)|^2
        $$

        where $G$ is a subset of the reciprocal space vectors of the supercell.
        """
        if self.verbose:
            print(f'Processing k-point {k0[0]:8.4f} {k0[1]:8.4f} {k0[2]:8.4f}')

        # find the K0 onto which k0 folds
        # k0 = G0 + K0
        K0, G0 = find_K_from_k(k0, self.M)
        # find index of K0
        ikpt = self.find_K_index(K0)

        # get the overlap G-vectors
        Gvalid, Gall = self.get_ovlap_G(ikpt=ikpt)
        # Gnew = Gvalid + k0 - K0
        Goffset = Gvalid + G0[np.newaxis, :]

        # Index of the Gvalid in 3D grid
        GallIndex = Gall % self.wfc._ngrid[np.newaxis, :]
        GoffsetIndex = Goffset % self.wfc._ngrid[np.newaxis, :]

        # 3d grid for planewave coefficients
        wfc_k_3D = np.zeros(self.wfc._ngrid, dtype=np.complex128)

        # the weights and corresponding energies
        P_Km = np.zeros(self.wfc._nbands, dtype=float)
        E_Km = np.zeros(self.wfc._nbands, dtype=float)

        for nb in range(self.wfc._nbands):
            # initialize the array to zero, which is unnecessary since the
            # GallIndex is the same for the same K-point
            # wfc_k_3D[:,:,:] = 0.0

            if self._lsoc:
                # pad the coefficients to 3D grid
                band_coeff = self.wfc.read_band_coeffs(ispin=whichspin, ikpt=ikpt, iband=nb + 1, norm=True)
                nplw = band_coeff.shape[0] // 2
                band_spinor_coeff = [band_coeff[:nplw], band_coeff[nplw:]]

                # energy
                E_Km[nb] = self.bands[whichspin - 1, ikpt - 1, nb]
                for Ispinor in range(2):
                    # band = band_spinor_coeff[Ispinor]
                    # band /= np.linalg.norm(band)
                    wfc_k_3D[GallIndex[:, 0], GallIndex[:, 1], GallIndex[:, 2]] = band_spinor_coeff[Ispinor]

                    # spectral weight
                    P_Km[nb] += np.linalg.norm(wfc_k_3D[GoffsetIndex[:, 0], GoffsetIndex[:, 1], GoffsetIndex[:, 2]])**2
            else:
                # pad the coefficients to 3D grid
                band_coeff = self.wfc.read_band_coeffs(ispin=whichspin, ikpt=ikpt, iband=nb + 1, norm=True)
                if self._lgam:
                    nplw = band_coeff.size
                    tmp = np.zeros((nplw * 2 - 1), dtype=band_coeff.dtype)
                    # for Gamma version, the coefficients corresponding to G \ne 0
                    # is multiplied by a factor of sqrt(2)
                    band_coeff[1:] /= np.sqrt(2.)
                    tmp[:nplw] = band_coeff
                    tmp[nplw:] = band_coeff[1:].conj()
                    band_coeff = tmp

                wfc_k_3D[GallIndex[:, 0], GallIndex[:, 1], GallIndex[:, 2]] = band_coeff
                # energy
                E_Km[nb] = self.bands[whichspin - 1, ikpt - 1, nb]
                # spectral weight
                P_Km[nb] = np.linalg.norm(wfc_k_3D[GoffsetIndex[:, 0], GoffsetIndex[:, 1], GoffsetIndex[:, 2]])**2

        return np.array((E_Km, P_Km), dtype=float).T

    def spectral_weight(self, kpoints: List):
        """
        Calculate the spectral weight for a list of kpoints in the primitive BZ.
        """

        NKPTS = len(kpoints)

        sw = []
        for ispin in range(self.wfc._nspin):
            if self.wfc._nspin == 2:
                if self.verbose:
                    print('#' * 60)
                    print(f'Spin component: {ispin}')
                    print('#' * 60)
            sw.append([self.spectral_weight_k(kpoints[ik], whichspin=ispin + 1) for ik in range(NKPTS)])

        self.SW = np.array(sw)

        return self.SW

    def spectral_function(self, nedos: int = 4000, sigma: float = 0.02):
        r"""
        Generate the spectral function

        $$
            A(k_i, E) = \sum_m P_{Km}(k_i)\Delta(E - Em)
        $$

        Where the $\Delta$ function can be approximated by Lorentzian or Gaussian
        function.
        """

        assert self.SW is not None, 'Spectral weight must be calculated first!'

        NS = self.wfc._nspin
        # Number of kpoints
        nk = self.SW.shape[1]
        # spectral function
        SF = np.zeros((NS, nedos, nk), dtype=float)

        emin = self.SW[:, :, :, 0].min()
        emax = self.SW[:, :, :, 0].max()
        e0 = np.linspace(emin - 5 * sigma, emax + 5 * sigma, nedos)

        for ispin in range(NS):
            for ii in range(nk):
                E_Km = self.SW[ispin, ii, :, 0]
                P_Km = self.SW[ispin, ii, :, 1]

                SF[ispin, :, ii] = np.sum(LorentzSmearing(e0[:, np.newaxis], E_Km[np.newaxis, :], sigma=sigma) * P_Km[np.newaxis, :],
                                          axis=1)
        return e0, SF


def spectral_weight_multiple_source(kpoints: list, unfold_objs: List[Unfold], transform_matrix: np.ndarray):
    """
    Calculate the spectral weight for a list of kpoints in the primitive BZ
    from a list of `WAVECARs`.
    """

    nk = len(kpoints)
    ns = unfold_objs[0].wfc._nspin
    for obj in unfold_objs[1:]:
        assert ns == obj.wfc._nspin

    # When reading from multiple WAVECAR, it is possible that each of them may have differnt number of bands
    # if so, we take only the first N bands, where N is the minimum values of bands
    nb = []
    for source in unfold_objs:
        nb.append(source.bands.shape[2])
    nbands = min(nb)

    spectral_weights = []
    for ispin in range(ns):
        sw_this_spin = []
        for ik in range(nk):
            this_k = kpoints[ik]
            this_k_supercell, _ = find_K_from_k(this_k, transform_matrix)
            # find index of K0
            ok = False
            for source in unfold_objs:
                try:
                    source.find_K_index(this_k_supercell)
                except ValueError:
                    continue
                sw_this_spin.append(source.spectral_weight_k(this_k, whichspin=ispin + 1)[:nbands, :])
                ok = True
                break

            if not ok:
                raise ValueError(f'This kpoint {this_k} this not found in the WAVECAR(s) provided')

        spectral_weights.append(sw_this_spin)

    return np.array(spectral_weights)


def concatenate_scf_kpoints(scf_kpts: list, scf_weights: list, kpoints: list):
    """
    Concatenate SCF kpoints (from IBZKPT) with zero-weighted kpoints
    """
    scf_kpts = np.asarray(scf_kpts)
    kpoints = np.concatenate([scf_kpts, kpoints], axis=0)
    weights = np.zeros(kpoints.shape[0])
    weights[:len(scf_weights)] = scf_weights
    return kpoints, weights


def create_white_colormap(color: Union[str, tuple, list]) -> ListedColormap:
    """
    Create colormap from white to certain colour.

    Args:
        color (str, tuple, list): HEX color string or tuple/list of RGB color.
    """
    if isinstance(color, str):
        color = hex2color(color)
    color = color[:3]  # Ensure RGB
    N = 256
    vals = np.ones((N, 4))
    vals[:, 0] = np.linspace(1, color[0], N)
    vals[:, 1] = np.linspace(1, color[1], N)
    vals[:, 2] = np.linspace(1, color[2], N)
    return ListedColormap(vals)


def create_white_colormap_from_existing(name: str) -> ListedColormap:
    """
    Create a white-based color map from an existing one.
    """
    if name.startswith('#'):
        # Passed a HEX color, use it as the terminal colour as it is
        return create_white_colormap(name)

    cmp = plt.get_cmap(name)
    rgba = cmp(cmp.N)
    return create_white_colormap(rgba)


def parse_atoms_idx(atoms_idx: str) -> List[int]:
    """
    Expanding syntex like `1-2` (inclusive)

    For example, `1,2,3,4-6` will be expanded as `1,2,3,4,5,6`.

    :param atoms_idx: A string encode atom indices

    :returns: A list of atom indices
    """
    items = re.split(', *', atoms_idx)
    out = []
    for item in items:
        match = re.match(r'(\d+)-(\d+)', item)
        if match:
            out.extend(range(int(match.group(1)), int(match.group(2)) + 1))
        else:
            out.append(int(item))
    # Expect passing 1-based indexing
    out = [x - 1 for x in out]
    return out


def process_projection_options(atoms_idx: str, orbitals: str) -> Tuple[list, list]:
    """
    Process commandline-style specifications for project

    :returns: A tuple of atom indices and the orbitals selected for projection.
    """
    indices = parse_atoms_idx(atoms_idx)
    if orbitals and orbitals != 'all':
        orbitals = [token.strip() for token in orbitals.split(',')]
    else:
        orbitals = 'all'
    return indices, orbitals
